package com.devbeh;


import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.stream.Collectors;

import com.codahale.metrics.MetricRegistry;
import com.devbeh.kafka.KafkaUtils;
import com.devbeh.managers.ActionsManager;
import com.devbeh.managers.DetectionManager;
import com.devbeh.managers.DetectionsVisualizationManager;
import com.devbeh.managers.DistanceEstimationManager;
import com.devbeh.managers.KBManager;
import com.devbeh.managers.PerceptionVisualizationManager;
import com.devbeh.managers.PoseEstimationManager;
import com.devbeh.managers.TrackingManager;
import com.devbeh.managers.VideoProvisionManager;
import com.devbeh.videoproviders.FileVideoProvider;
import org.apache.kafka.clients.admin.AdminClient;
import org.apache.kafka.clients.admin.Config;
import org.apache.kafka.clients.admin.ConfigEntry;
import org.apache.kafka.clients.admin.CreateTopicsResult;
import org.apache.kafka.clients.admin.NewPartitions;
import org.apache.kafka.clients.admin.NewTopic;
import org.apache.kafka.common.KafkaFuture;
import org.apache.kafka.common.Node;
import org.apache.kafka.common.config.ConfigResource;
import org.apache.kafka.common.errors.InvalidPartitionsException;
import org.apache.kafka.common.errors.TopicExistsException;
import org.apache.kafka.common.errors.UnknownTopicOrPartitionException;
import org.springframework.beans.factory.DisposableBean;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.SpringApplication;
import org.springframework.boot.autoconfigure.SpringBootApplication;
import org.springframework.boot.context.properties.ConfigurationProperties;
import org.springframework.context.ApplicationContext;
import org.springframework.context.annotation.Bean;
import org.springframework.core.env.Environment;


@SuppressWarnings("SpringJavaAutowiredFieldsWarningInspection")
@SpringBootApplication
public class DeviantBehaviourManagerStarter implements InitializingBean, DisposableBean
{
    public static void main(String[] args)
    {
        SpringApplication.run(DeviantBehaviourManagerStarter.class, args);
    }


    /**
     * Содержит функции для закрытия pipeline'ов при завершении работы.
     */
    private final List<Runnable> pipelineStoppers = new ArrayList<>();

    @Autowired
    private MetricRegistry metricRegistry;

    @Autowired
    Environment env;

    @Autowired
    private ApplicationContext ctx;


    @ConfigurationProperties("dbm")
    @Bean
    public DbmProperties dbmProperties()
    {
        return new DbmProperties();
    }


    private void validateAndFixProperties(DbmProperties prop)
    {
        checkCondition(!Utils.isEmpty(prop.getBootstrapServer()), "dbm.bootstrap-server is missing");
        checkCondition(!Utils.isEmpty(prop.getDetectionRequestTopicName()), "dbm.detection-request-topic-name must not be empty");
        checkCondition(!Utils.isEmpty(prop.getDetectionResponseTopicPrefix()), "dbm.detection-response-topic-prefix must not be empty");
        checkCondition(prop.getMaxProcessingFrames() > 1, "dbm.max-processing-frames must be greater than 1");
        checkCondition(prop.getMaxFrameSizePixels() > 0, "dbm.max-frame-size-pixels must be positive");
        checkCondition(prop.getFileVideoProviders() != null && !prop.getFileVideoProviders().isEmpty(),
                "No dbm.file-video-providers specified");
        checkCondition(!Utils.isEmpty(prop.getTrackingRequestTopicName()), "dbm.tracking-request-topic-name must not be empty");
        checkCondition(!Utils.isEmpty(prop.getTrackingResponseTopicPrefix()), "dbm.tracking-response-topic-prefix must not be empty");
        checkCondition(!Utils.isEmpty(prop.getTrackingManagerStateTopicPrefix()), "dbm.tracking-manager-state-topic-prefix must not be empty");
        checkCondition(!Utils.isEmpty(prop.getPoseRequestTopicName()), "dbm.pose-request-topic-name must not be empty");
        checkCondition(!Utils.isEmpty(prop.getPoseResponseTopicPrefix()), "dbm.pose-response-topic-prefix must not be empty");
        checkCondition(!Utils.isEmpty(prop.getDistanceRequestTopicName()), "dbm.distance-request-topic-name must not be empty");
        checkCondition(!Utils.isEmpty(prop.getDetectionResponseTopicPrefix()), "dbm.distance-response-topic-prefix must not be empty");
        checkCondition(!Utils.isEmpty(prop.getActionsRequestTopicName()), "dbm.actions-request-topic-name must not be empty");
        checkCondition(!Utils.isEmpty(prop.getActionsResponseTopicPrefix()), "dbm.actions-response-topic-prefix must not be empty");
        checkCondition(!Utils.isEmpty(prop.getKbStateTopicPrefix()), "dbm.kb-state-topic-prefix must not be empty");
        checkCondition(!Utils.isEmpty(prop.getDevBehDetectionsTopicPrefix()), "dbm.dev-beh-detections-topic-prefix must not be empty");
        checkCondition(!Utils.isEmpty(prop.getDevBehDetectionsDirPath()), "dbm.dev-beh-detections-dir-path must not be empty");
        checkCondition(!Utils.isEmpty(prop.getPerceptionResultDirPath()), "dbm.perception-result-dir-path must not be empty");
        checkCondition(prop.getBoxContactPointLiftPercent() >= 0 && prop.getBoxContactPointLiftPercent() <= 1,
                "dbm.box-contact-point-lift-percent must be in the [0,1] interval");
        checkCondition(prop.getInferencePeriodSec() >= 0, "dbm.inference-period-sec must not be negative");
        checkCondition(prop.getDevBehIgnorePeriodSec() > 0, "dbm.dev-beh-ignore-period-sec must be positive");
        checkCondition(!Utils.isEmpty(prop.getKbWorkerJarPath()), "dbm.kb-worker-jar-path must not be empty");
        checkCondition(!Utils.isEmpty(prop.getKbWorkerMemory()), "dbm.kb-worker-memory must not be empty");
        checkCondition(!Utils.isEmpty(prop.getLibjplJavaLibraryPath()), "dbm.libjpl-java-library-path must not be empty");
        checkCondition(!Utils.isEmpty(prop.getGeneralKbRulesFilePath()), "dbm.general-kb-rules-file-path must not be empty");
        checkCondition(prop.getMaxActionDurationSec() > 0, "dbm.max-action-duration-sec must be positive");
        checkCondition(prop.getMaxChildSkeletonLengthMm() > 0, "dbm.max-child-skeleton-length-mm must be positive");
        checkCondition(prop.getMaxOccludedDurationSec() > 0, "dbm.max-occluded-duration-sec must be positive");
        checkCondition(prop.getKbHistoryIntervalSec() > 1, "dbm.kb-history-interval-sec must be greater than 1");
        checkCondition(prop.getMaxQueuedFrames() > 1, "dbm.max-queued-frames must be greater than 1");
        checkCondition(prop.getMaxProcessingFrames() > 1, "dbm.max-processing-frames must be greater than 1");

        File devBehDetectionsDirFile = new File(prop.getDevBehDetectionsDirPath());
        checkCondition(!devBehDetectionsDirFile.isFile(), "The devBehDetectionsDir '" +
                prop.getDevBehDetectionsDirPath() + "' should be a directory");

        File perceptionResultDirFile = new File(prop.getPerceptionResultDirPath());
        checkCondition(!perceptionResultDirFile.isFile(), "The perceptionResultDir '" +
                prop.getPerceptionResultDirPath() + "' should be a directory");

        File generalKbRulesFile = new File(prop.getGeneralKbRulesFilePath());
        checkCondition(generalKbRulesFile.isFile() && generalKbRulesFile.canRead(),
                "General kb rules file '" + prop.getGeneralKbRulesFilePath() + "' can't be read");

        Set<String> videoStreamNames = new HashSet<>();
        boolean manualPartition = false;
        boolean automaticPartition = false;
        for (DbmProperties.FileVideoProviderProperties fvpProp : prop.getFileVideoProviders())
        {
            checkCondition(fvpProp.getVideoId() >= -1, "dbm.file-video-providers[].video-id "
                    + "must be specified");

            if (fvpProp.getVideoId() == -1)
            {
                automaticPartition = true;
            }
            else if (fvpProp.getVideoId() >= 0)
            {
                manualPartition = true;
            }

            checkCondition(!Utils.isEmpty(fvpProp.getName()), "dbm.file-video-providers[].name must be specified");
            checkCondition(videoStreamNames.add(fvpProp.getName()),
                    "dbm.file-video-providers[].name must be unique, but got duplicate " + fvpProp.getName());
            checkCondition(!Utils.isEmpty(fvpProp.getFilePath()), "dbm.file-video-providers[].file-path shouldn't be empty");
            checkCondition(fvpProp.getMaxFps() > 0 || prop.getMaxFps() > 0,
                    "Either dbm.file-video-providers[].max-fps or dbm.max-fps must be specified (with positive value)");
            checkCondition(fvpProp.getCameraHeightMm() > 0, "dbm.file-video-providers[].camera-height-mm must be positive");
            checkCondition(fvpProp.getFocalLengthMm() > 0, "dbm.file-video-providers[].focal-length-mm must be positive");
            checkCondition(fvpProp.getPixelSizeMm() > 0, "dbm.file-video-providers[].pixel-size-mm must be positive");
            if (!Utils.isEmpty(fvpProp.getCameraKbRulesFilePath()))
            {
                File cameraKbRulesFile = new File(fvpProp.getCameraKbRulesFilePath());
                checkCondition(cameraKbRulesFile.isFile() && cameraKbRulesFile.canRead(),
                        "Camera kb rules file '" + fvpProp.getCameraKbRulesFilePath() + "' can't be read");
            }

            fvpProp.setMaxFps(fvpProp.getMaxFps() > 0 ? fvpProp.getMaxFps() : prop.getMaxFps());
        }

        checkCondition(!(manualPartition && automaticPartition), "All video id's must be either specified "
                + "manually or assigned automatically");

        if (automaticPartition)
        {
            for (int i = 0; i < prop.getFileVideoProviders().size(); i++)
            {
                prop.getFileVideoProviders().get(i).setVideoId(i);
            }
        }
    }


    private static void checkCondition(boolean condition, String errorMsg)
    {
        if (!condition)
        {
            throw new RuntimeException(errorMsg);
        }
    }


    @Override
    public void afterPropertiesSet() throws Exception
    {
        DbmProperties prop = dbmProperties();
        validateAndFixProperties(prop);

        // Создаем необходимые партиции и очищаем данные KB, при необходимости

        int partitionsCount = prop.getFileVideoProviders().stream()
                .mapToInt(DbmProperties.FileVideoProviderProperties::getVideoId).max().orElse(-1) + 1;
        if (partitionsCount == 0)
        {
            return;
        }

        // Список топиков для отправки запросов в компоненты анализа
        List<String> requestTopics = List.of(prop.getDetectionRequestTopicName(), prop.getTrackingRequestTopicName(),
                prop.getPoseRequestTopicName(), prop.getDistanceRequestTopicName(), prop.getActionsRequestTopicName());
        try(AdminClient ac = KafkaUtils.createAdminClient(prop.getBootstrapServer(), prop.getUsername(),
                prop.getPassword()))
        {
            final Collection<Node> nodes = ac.describeCluster().nodes().get();
            if (nodes.isEmpty())
            {
                throw new RuntimeException("No brokers available to obtain default settings");
            }

            String nodeId = nodes.iterator().next().idString();
            ConfigResource cr = new ConfigResource(ConfigResource.Type.BROKER, nodeId);
            Map<ConfigResource, Config> configs = ac.describeConfigs(Collections.singleton(cr)).all().get();
            Config brokerConfig = configs.get(cr);
            if (brokerConfig == null)
            {
                throw new RuntimeException("Can't access broker config");
            }

            ConfigEntry replicationFactorCE = brokerConfig.get("default.replication.factor");
            if (replicationFactorCE == null)
            {
                throw new RuntimeException("Can't access default.replication.factor");
            }

            short replicationFactor = Short.parseShort(replicationFactorCE.value());

            CreateTopicsResult ctr = ac.createTopics(requestTopics.stream()
                    .map(t -> new NewTopic(t, partitionsCount, replicationFactor)).collect(Collectors.toSet()));

            Map<String, NewPartitions> partitionsToCreate = new HashMap<>();
            for (Map.Entry<String, KafkaFuture<Void>> ctrEntry : ctr.values().entrySet())
            {
                try
                {
                    ctrEntry.getValue().get();
                }
                catch (ExecutionException e)
                {
                    if (!(e.getCause() instanceof TopicExistsException))
                    {
                        throw e;
                    }

                    // Топик уже существовал
                    partitionsToCreate.put(ctrEntry.getKey(), NewPartitions.increaseTo(partitionsCount));
                }
            }

            if (!partitionsToCreate.isEmpty())
            {
                try
                {
                    ac.createPartitions(partitionsToCreate).all().get();
                }
                catch (ExecutionException e)
                {
                    // InvalidPartitionsException - если топик уже имеет достаточное количество партиций
                    if (!(e.getCause() instanceof InvalidPartitionsException))
                    {
                        throw e;
                    }
                }
            }

            // При необходимости, удаляем данные KB
            Set<String> topicsToDelete = new HashSet<>();
            for (DbmProperties.FileVideoProviderProperties fvpProp : prop.getFileVideoProviders())
            {
                if (fvpProp.isClearKbState())
                {
                    topicsToDelete.add(prop.getKbStateTopicPrefix() + fvpProp.getVideoId());
                    topicsToDelete.add(prop.getDevBehDetectionsTopicPrefix() + fvpProp.getVideoId());
                }
            }

            if (!topicsToDelete.isEmpty())
            {
                try
                {
                    ac.deleteTopics(topicsToDelete).all().get();
                }
                catch (ExecutionException e)
                {
                    if (!(e.getCause() instanceof UnknownTopicOrPartitionException))
                    {
                        throw e;
                    }
                }
            }
        }

        // Инициализируем конвейеры обработки видеозаписей

        List<Actor> managers = new ArrayList<>();
        MetricService metricService = new MetricService(metricRegistry);

        for (DbmProperties.FileVideoProviderProperties fvpProp : prop.getFileVideoProviders())
        {
            // Теоретически, если два потока одновременно запустят остановку pipeline, они могут заджойниться друг на
            // друге, но считаем, что это маловероятно, и если такое возникнет, сработает ограничение максимального
            // времени ожидания
            PipelineStopper pipelineStopper = new PipelineStopper();

            CompletableFuture<VideoInfo> fpsFuture = new CompletableFuture<>();

            // Создаем компонент визуализации результатов детектирования
            FileVideoProvider dvmVideoProvider = new FileVideoProvider(fvpProp.getFilePath(), fvpProp.getMaxFps());
            DetectionsVisualizationManager detectionsVisualizationManager = new DetectionsVisualizationManager(
                    fvpProp.getName(), fvpProp.getVideoId(), prop.getBootstrapServer(), prop.getUsername(),
                    prop.getPassword(), prop.getDevBehDetectionsTopicPrefix(), prop.getDevBehDetectionsDirPath(),
                    prop.getDevBehDetectionsStartFrameId(), dvmVideoProvider, prop.getMaxFrameSizePixels(),
                    pipelineStopper::stop, metricService);

            // Создание менеджера визуализации результатов анализа видеозаписей
            PerceptionVisualizationManager perceptionVisualizationManager = new PerceptionVisualizationManager(
                    fvpProp.getName(), prop.getPerceptionResultDirPath(), prop.getPerceptionInitialConsumedFrameId(),
                    prop.isPerceptionVisualizationEnabled(), prop.getMaxQueuedFrames(), fpsFuture,
                    pipelineStopper::stop, metricService);

            // Создание менеджера базы знаний
            KBManager kbManager = new KBManager(fvpProp.getName(), fvpProp.getVideoId(), prop.getBootstrapServer(),
                    prop.getUsername(), prop.getPassword(), prop.getKbStateTopicPrefix(),
                    prop.getDevBehDetectionsTopicPrefix(), prop.getKbHistoryIntervalSec(), prop.getInferencePeriodSec(),
                    prop.getDevBehIgnorePeriodSec(), fpsFuture, prop.getMaxQueuedFrames(), prop.getKbWorkerJarPath(),
                    prop.getKbWorkerMemory(), prop.getLibjplJavaLibraryPath(), prop.getGeneralKbRulesFilePath(),
                    fvpProp.getCameraKbRulesFilePath(), prop.getAtomicKnowledgeDirPath(),
                    prop.getMaxActionDurationSec(), prop.getMaxChildSkeletonLengthMm(), pipelineStopper::stop,
                    metricService);

            // Создание менеджера детектирования действий
            ActionsManager actionsManager = new ActionsManager(fvpProp.getName(), fvpProp.getVideoId(),
                    prop.getBootstrapServer(), prop.getUsername(), prop.getPassword(),
                    prop.getActionsRequestTopicName(), prop.getActionsResponseTopicPrefix(),  prop.getMaxQueuedFrames(),
                    prop.getMaxProcessingFrames(), kbManager, perceptionVisualizationManager, pipelineStopper::stop,
                    metricService);

            // Создание менеджера оценки расстояния до объектов
            DistanceEstimationManager distanceEstimationManager = new DistanceEstimationManager(fvpProp.getName(),
                    fvpProp.getVideoId(), fvpProp.getCameraHeightMm(), fvpProp.getFocalLengthMm(),
                    fvpProp.getPixelSizeMm(), fvpProp.getNormal(), prop.getBootstrapServer(), prop.getUsername(),
                    prop.getPassword(), prop.getDistanceRequestTopicName(), prop.getDistanceResponseTopicPrefix(),
                    prop.getBoxContactPointLiftPercent(), prop.getMaxQueuedFrames(), prop.getMaxProcessingFrames(),
                    actionsManager, kbManager, perceptionVisualizationManager, pipelineStopper::stop, metricService);

            // Создание менеджера оценки трехмерного скелета людей
            PoseEstimationManager poseEstimationManager = new PoseEstimationManager(fvpProp.getName(),
                    fvpProp.getVideoId(), prop.getBootstrapServer(), prop.getUsername(), prop.getPassword(),
                    prop.getPoseRequestTopicName(), prop.getPoseResponseTopicPrefix(), prop.getMaxQueuedFrames(),
                    prop.getMaxProcessingFrames(), distanceEstimationManager, perceptionVisualizationManager,
                    pipelineStopper::stop, metricService);

            // Создание менеджера отслеживания людей и объектов
            TrackingManager trackingManager = new TrackingManager(fvpProp.getName(), fvpProp.getVideoId(),
                    prop.getBootstrapServer(), prop.getUsername(), prop.getPassword(),
                    prop.getTrackingRequestTopicName(), prop.getTrackingResponseTopicPrefix(),
                    prop.getTrackingManagerStateTopicPrefix(), prop.getMaxOccludedDurationSec(), fpsFuture,
                    prop.getMaxQueuedFrames(), prop.getMaxProcessingFrames(), distanceEstimationManager,
                    poseEstimationManager, actionsManager, kbManager, perceptionVisualizationManager,
                    pipelineStopper::stop, metricService);

            // Создание менеджера детектирования людей и объектов
            DetectionManager detectionManager = new DetectionManager(fvpProp.getName(), fvpProp.getVideoId(),
                    prop.getBootstrapServer(), prop.getUsername(), prop.getPassword(),
                    prop.getDetectionRequestTopicName(), prop.getDetectionResponseTopicPrefix(),
                    prop.getMaxQueuedFrames(), prop.getMaxProcessingFrames(), trackingManager, poseEstimationManager,
                    pipelineStopper::stop, metricService);

            // Создание менеджера предоставления кадров
            FileVideoProvider videoProvider = new FileVideoProvider(fvpProp.getFilePath(), fvpProp.getMaxFps());
            VideoProvisionManager videoProvisionManager = new VideoProvisionManager(fvpProp.getName(), videoProvider,
                    prop.getMaxFrameSizePixels(), fpsFuture, detectionManager, trackingManager, poseEstimationManager,
                    distanceEstimationManager, perceptionVisualizationManager, pipelineStopper::stop, metricService);

            // Инициализация объекта для остановки обработки видеозаписи
            pipelineStopper.init(videoProvisionManager, detectionManager, trackingManager, poseEstimationManager,
                    distanceEstimationManager, actionsManager, kbManager, perceptionVisualizationManager,
                    detectionsVisualizationManager);
            pipelineStoppers.add(pipelineStopper::stop);

            // Запуск менеджеров обработки видео
            startManager(videoProvisionManager, managers);
            startManager(detectionManager, managers);
            startManager(trackingManager, managers);
            startManager(poseEstimationManager, managers);
            startManager(distanceEstimationManager, managers);
            startManager(actionsManager, managers);
            startManager(kbManager, managers);
            startManager(perceptionVisualizationManager, managers);
            startManager(detectionsVisualizationManager, managers);
        }

        // Запускаем ожидание завершения работы и выключаем систему
        Thread stopperThread = new Thread(() -> {
            try
            {
                for (Actor m : managers)
                {
                    m.joinThread();
                }

                SpringApplication.exit(ctx, () -> 0);
            }
            catch (InterruptedException ignore)
            {
                // Может быть выброшен только при выключении системы
            }
        });
        stopperThread.setName("ServiceStopper");
        stopperThread.setDaemon(false);
        stopperThread.start();
    }


    private static void startManager(Actor actor, List<Actor> actors)
    {
        actor.start();
        actors.add(actor);
    }


    @Override
    public void destroy()
    {
        pipelineStoppers.forEach(Runnable::run);
    }


    private static class PipelineStopper
    {
        private VideoProvisionManager videoProvisionManager;
        private DetectionManager detectionManager;
        private TrackingManager trackingManager;
        private PoseEstimationManager poseEstimationManager;
        private DistanceEstimationManager distanceEstimationManager;
        private ActionsManager actionsManager;
        private KBManager kbManager;
        private PerceptionVisualizationManager perceptionVisualizationManager;
        private DetectionsVisualizationManager detectionsVisualizationManager;


        public void init(VideoProvisionManager videoProvisionManager, DetectionManager detectionManager,
                TrackingManager trackingManager, PoseEstimationManager poseEstimationManager,
                DistanceEstimationManager distanceEstimationManager, ActionsManager actionsManager,
                KBManager kbManager, PerceptionVisualizationManager perceptionVisualizationManager,
                DetectionsVisualizationManager detectionsVisualizationManager)
        {
            this.videoProvisionManager = videoProvisionManager;
            this.detectionManager = detectionManager;
            this.trackingManager = trackingManager;
            this.poseEstimationManager = poseEstimationManager;
            this.distanceEstimationManager = distanceEstimationManager;
            this.actionsManager = actionsManager;
            this.kbManager = kbManager;
            this.perceptionVisualizationManager = perceptionVisualizationManager;
            this.detectionsVisualizationManager = detectionsVisualizationManager;
        }


        public void stop()
        {
            this.videoProvisionManager.stop();
            this.detectionManager.stop();
            this.trackingManager.stop();
            this.poseEstimationManager.stop();
            this.distanceEstimationManager.start();
            this.actionsManager.stop();
            this.kbManager.stop();
            this.perceptionVisualizationManager.stop();
            this.detectionsVisualizationManager.stop();
        }
    }
}
