package com.devbeh;


import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import com.codahale.metrics.Counter;
import com.codahale.metrics.Gauge;
import com.codahale.metrics.Meter;
import com.codahale.metrics.MetricRegistry;
import org.springframework.boot.actuate.metrics.CounterService;
import org.springframework.boot.actuate.metrics.GaugeService;


public class MetricService implements CounterService, GaugeService
{
    private final MetricRegistry registry;

    private final Map<String, SimpleGauge> gauges = new ConcurrentHashMap<>();

    private final Map<String, String> names = new ConcurrentHashMap<>();

    /**
     * @param registry the underlying metric registry
     */
    public MetricService(MetricRegistry registry)
    {
        this.registry = registry;
    }


    @Override
    public void increment(String name)
    {
        incrementInternal(name, 1L);
    }


    @Override
    public void decrement(String name)
    {
        incrementInternal(name, -1L);
    }


    private void incrementInternal(String name, long value)
    {
        if (name.startsWith("meter"))
        {
            Meter meter = this.registry.meter(name);
            meter.mark(value);
        }
        else
        {
            name = wrapCounterName(name);
            Counter counter = this.registry.counter(name);
            counter.inc(value);
        }
    }


    @Override
    public void submit(String name, double value)
    {
        name = wrapGaugeName(name);
        setGaugeValue(name, value);
    }


    private void setGaugeValue(String name, double value)
    {
        SimpleGauge gauge = this.gauges.get(name);
        if (gauge == null)
        {
            SimpleGauge newGauge = new SimpleGauge(value);
            gauge = this.gauges.putIfAbsent(name, newGauge);
            if (gauge == null)
            {
                this.registry.register(name, newGauge);
                return;
            }
        }
        gauge.setValue(value);
    }


    private String wrapGaugeName(String metricName)
    {
        return wrapName(metricName, "gauge.");
    }


    private String wrapCounterName(String metricName)
    {
        return wrapName(metricName, "counter.");
    }


    private String wrapName(String metricName, String prefix)
    {
        String cached = this.names.get(metricName);
        if (cached != null)
        {
            return cached;
        }
        if (metricName.startsWith(prefix))
        {
            return metricName;
        }
        String name = prefix + metricName;
        this.names.put(metricName, name);
        return name;
    }


    @Override
    public void reset(String name)
    {
        this.registry.remove(name);
        names.remove(name);
    }

    /**
     * Simple {@link Gauge} implementation to {@literal double} value.
     */
    private static final class SimpleGauge implements Gauge<Double>
    {

        private volatile double value;

        private SimpleGauge(double value)
        {
            this.value = value;
        }


        @Override
        public Double getValue()
        {
            return this.value;
        }


        public void setValue(double value)
        {
            this.value = value;
        }

    }
}

