package com.devbeh.kafka;


import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import com.devbeh.Utils;
import org.apache.kafka.clients.CommonClientConfigs;
import org.apache.kafka.clients.admin.AdminClient;
import org.apache.kafka.clients.admin.AdminClientConfig;
import org.apache.kafka.clients.consumer.ConsumerConfig;
import org.apache.kafka.clients.consumer.KafkaConsumer;
import org.apache.kafka.clients.producer.KafkaProducer;
import org.apache.kafka.clients.producer.ProducerConfig;
import org.apache.kafka.common.TopicPartition;
import org.apache.kafka.common.config.SaslConfigs;
import org.apache.kafka.common.security.auth.SecurityProtocol;
import org.apache.kafka.common.security.plain.PlainLoginModule;
import org.apache.kafka.common.security.plain.internals.PlainSaslServer;
import org.apache.kafka.common.serialization.Deserializer;
import org.apache.kafka.common.serialization.LongDeserializer;
import org.apache.kafka.common.serialization.Serializer;


public class KafkaUtils
{
    /**
     * Создает экземпляр {@link AdminClient}.
     *
     * @param bootstrapServer сервер Кафки
     * @param username имя пользователя для подключения по SASL. Если username или password - пустые строки или null,
     *                 то используется обычное подключение
     * @param password пароль для подключения по SASL
     */
    public static AdminClient createAdminClient(String bootstrapServer, String username, String password)
    {
        Map<String, Object> conf = new HashMap<>();
        conf.put(AdminClientConfig.BOOTSTRAP_SERVERS_CONFIG, bootstrapServer);
        conf.put(AdminClientConfig.REQUEST_TIMEOUT_MS_CONFIG, "60000");
        // Устанавливаем параметры аутентификации
        //noinspection DuplicatedCode
        if (!Utils.isEmpty(username) && !Utils.isEmpty(password))
        {
            conf.put(CommonClientConfigs.SECURITY_PROTOCOL_CONFIG, SecurityProtocol.SASL_PLAINTEXT.name);
            conf.put(SaslConfigs.SASL_MECHANISM, PlainSaslServer.PLAIN_MECHANISM);
            String saslJaasConfig = "org.apache.kafka.common.security.plain.PlainLoginModule required username=\""
                    + username + "\" password=\"" + password + "\";";
            conf.put(SaslConfigs.SASL_JAAS_CONFIG, saslJaasConfig);
        }
        return AdminClient.create(conf);
    }


    /**
     * Создает экземпляр {@link KafkaConsumer}.
     *
     * @param bootstrapServer сервер Кафки
     * @param username имя пользователя для подключения по SASL. Если username или password - пустые строки или null,
     *                 то используется обычное подключение
     * @param password пароль для подключения по SASL
     * @param groupId группа клиента Кафки
     * @param clientId идентификатор клиента Кафки
     * @param partition партиция, из которой читаются сообщения
     * @param valueDeserializer десериализатор значения
     * @return экземпляр {@link KafkaConsumer}
     */
    public static <T> KafkaConsumer<?, T> buildKafkaConsumer(String bootstrapServer, String username, String password,
            String groupId, String clientId, TopicPartition partition, Deserializer<T> valueDeserializer)
    {
        Properties p = new Properties();
        p.put(ConsumerConfig.BOOTSTRAP_SERVERS_CONFIG, bootstrapServer);
        p.put(ConsumerConfig.ENABLE_AUTO_COMMIT_CONFIG, false);
        p.put(ConsumerConfig.AUTO_OFFSET_RESET_CONFIG, "earliest");
        // Уменьшает время poll при запросе старых данных
        p.put(ConsumerConfig.MAX_PARTITION_FETCH_BYTES_CONFIG, "5000000");
        p.put(ConsumerConfig.GROUP_ID_CONFIG, groupId);
        p.put(ConsumerConfig.CLIENT_ID_CONFIG, clientId);

        // Устанавливаем параметры аутентификации
        if (!Utils.isEmpty(username) && !Utils.isEmpty(password))
        {
            p.put(CommonClientConfigs.SECURITY_PROTOCOL_CONFIG, SecurityProtocol.SASL_PLAINTEXT.name);
            p.put(SaslConfigs.SASL_MECHANISM, PlainSaslServer.PLAIN_MECHANISM);
            String saslJaasConfig = "org.apache.kafka.common.security.plain.PlainLoginModule required username=\""
                    + username + "\" password=\"" + password + "\";";
            p.put(SaslConfigs.SASL_JAAS_CONFIG, saslJaasConfig);
        }

        KafkaConsumer<?, T> consumer = new KafkaConsumer<>(p, new LongDeserializer(), valueDeserializer);
        consumer.assign(Collections.singletonList(partition));

        return consumer;
    }


    /**
     * Создает экземпляр {@link KafkaProducer}.
     *
     * @param bootstrapServer сервер Кафки
     * @param username имя пользователя для подключения по SASL. Если username или password - пустые строки или null,
     *                 то используется обычное подключение
     * @param password пароль для подключения по SASL
     * @param clientId идентификатор клиента Кафки
     * @param valueSerializer сериализатор значения
     * @param <T> тип значения
     * @return экземпляр {@link KafkaProducer}
     */
    public static <T> KafkaProducer<?, T> buildKafkaProducer(String bootstrapServer, String username, String password,
            String clientId, Serializer<T> valueSerializer)
    {
        Map<String, Object> config = new HashMap<>();

        config.put(ProducerConfig.CLIENT_ID_CONFIG, clientId);
        config.put(ProducerConfig.BOOTSTRAP_SERVERS_CONFIG, bootstrapServer);

        config.put(ProducerConfig.ACKS_CONFIG, "all");
        config.put(ProducerConfig.RETRIES_CONFIG, String.valueOf(Integer.MAX_VALUE));
        config.put(ProducerConfig.ENABLE_IDEMPOTENCE_CONFIG, "true");
        // max.block.ms остается по умолчанию и учитывается при отправке, чтобы не допустить потерю сообщений
        config.put(ProducerConfig.DELIVERY_TIMEOUT_MS_CONFIG, String.valueOf(Integer.MAX_VALUE));

        // Устанавливаем параметры аутентификации
        if (!Utils.isEmpty(username) && !Utils.isEmpty(password))
        {
            config.put(CommonClientConfigs.SECURITY_PROTOCOL_CONFIG, SecurityProtocol.SASL_PLAINTEXT.name);
            config.put(SaslConfigs.SASL_MECHANISM, PlainSaslServer.PLAIN_MECHANISM);
            String saslJaasConfig = PlainLoginModule.class.getName() + " required username=\""
                    + username + "\" password=\"" + password + "\";";
            config.put(SaslConfigs.SASL_JAAS_CONFIG, saslJaasConfig);
        }

        return new KafkaProducer<>(config, (topic, data) -> null, valueSerializer);
    }
}
