package com.devbeh.managers;


import java.time.Duration;
import java.util.Collections;
import java.util.Map;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.atomic.AtomicLong;

import com.devbeh.Actor;
import com.devbeh.MetricService;
import com.devbeh.ReducibleSemaphore;
import com.devbeh.data.PerceptionRequest;
import com.devbeh.data.PerceptionResponse;
import com.devbeh.kafka.KafkaUtils;
import com.devbeh.kafka.PerceptionRequestDeserializer;
import com.devbeh.kafka.PerceptionResponseDeserializer;
import org.apache.kafka.clients.consumer.ConsumerRecord;
import org.apache.kafka.clients.consumer.KafkaConsumer;
import org.apache.kafka.clients.consumer.OffsetAndTimestamp;
import org.apache.kafka.common.TopicPartition;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


/**
 * Абстрактный класс для дистрибьюторов результата анализа кадров потребителям.
 */
public abstract class AbstractDistributor extends Actor implements DataConsumer
{
    private static final Logger LOGGER = LoggerFactory.getLogger(AbstractDistributor.class);

    /**
     * Счетчик ошибок.
     */
    private final AtomicLong errorCounter = new AtomicLong();

    /**
     * Используется в качестве имени потока дистрибьютора.
     */
    private final String distributorName;

    /**
     * Адрес сервера Kafka.
     */
    private final String bootstrapServer;

    /**
     * Имя пользователя Kafka для SASL аутентификации или пустая строка/null.
     */
    private final String username;

    /**
     * Пароль пользователя Kafka для SASL аутентификации или пустая строка/null.
     */
    private final String password;

    /**
     * Идентификатор клиента (и группы) Kafka, используемый для чтения топика запросов.
     */
    private final String clientId;

    /**
     * Партиция для записи запросов на обработку кадров.
     */
    private final TopicPartition requestTopicPartition;

    /**
     * Партиция результатов обработки кадров.
     */
    private final TopicPartition responseTopicPartition;

    /**
     * После инициализации содержит идентификатор первого кадра, который требуется для обработки на данном
     * этапе pipeline.
     */
    private final CompletableFuture<Long> initialLastConsumedFrameIdFuture;

    /**
     * Максимальное количество одновременно обрабатываемых кадров (отправленных в Kafka без получения
     * результата анализа).
     */
    private final int maxProcessingRequests;

    /**
     * Семафор, используемый для ограничения количества одновременно обрабатываемых запросов.
     * Используется и в потоке дистрибьютора и потоке отправителя запросов на анализ.
     */
    private final ReducibleSemaphore processingRequestsSem;

    /**
     * Используется для остановки всего pipeline обработки видеозаписи в случае критической ошибки.
     */
    private final Runnable pipelineStopper;

    /**
     * Используется для отображения метрик на странице metrics.
     */
    private final MetricService metricService;

    /**
     * Имя метрики последнего обработанного идентификатора кадра.
     */
    private final String lastFrameIdMetricName;

    /**
     * Имя метрики количества ошибок.
     */
    private final String totalErrorsMetricName;

    /**
     * Используется для чтения результатов обработки кадров.
     */
    private KafkaConsumer<?, PerceptionResponse> respConsumer;

    /**
     * Идентификатор последнего кадра, результаты обработки которого были направлены потребителям. Используется
     * для предотвращения повторной отправки результата обработки потребителям.
     */
    protected long lastDistributedFrameId;


    public AbstractDistributor(String distributorName, String metricsPrefix, String bootstrapServer, String username,
            String password, String clientId, TopicPartition requestTopicPartition,
            TopicPartition responseTopicPartition, CompletableFuture<Long> initialLastConsumedFrameIdFuture,
            int maxProcessingRequests, ReducibleSemaphore processingRequestsSem, Runnable pipelineStopper,
            MetricService metricService)
    {
        this.distributorName = distributorName;
        this.bootstrapServer = bootstrapServer;
        this.username = username;
        this.password = password;
        this.clientId = clientId;
        this.requestTopicPartition = requestTopicPartition;
        // Используем videoId в качестве managerId
        this.responseTopicPartition = responseTopicPartition;
        this.initialLastConsumedFrameIdFuture = initialLastConsumedFrameIdFuture;
        this.maxProcessingRequests = maxProcessingRequests;
        this.processingRequestsSem = processingRequestsSem;
        this.pipelineStopper = pipelineStopper;
        this.metricService = metricService;
        this.lastFrameIdMetricName = metricsPrefix + "lastFrameId";
        this.totalErrorsMetricName = metricsPrefix + "totalErrors";

        metricService.submit(totalErrorsMetricName, 0);
    }


    @Override
    public long getInitialLastConsumedFrameId() throws InterruptedException
    {
        try
        {
            return initialLastConsumedFrameIdFuture.get();
        }
        catch (ExecutionException e)
        {
            // Shouldn't happen!
            throw new RuntimeException(e.getCause());
        }
    }


    @Override
    protected void doFirst()
    {
        try
        {
            doFirstInternal();
        }
        catch (Exception e)
        {
            if (!getDoStop())
            {
                handleCriticalError("A fatal exception occurred during manager initialization ["
                        + distributorName + "]. Stopping the hole pipeline.", e);
            }
        }
    }


    /**
     * Увеличивает счетчик ошибок и обновляет соответствущую метрику.
     */
    protected void incrementErrorCounter()
    {
        metricService.submit(totalErrorsMetricName, errorCounter.incrementAndGet());
    }


    /**
     * Обрабатывает критическую ошибку: логирует, увеличивает счетчик ошибок и останавливает pipeline обработки видео.
     *
     * @param errorMsg описание ошибки
     * @param e возникшее исключение или null
     */
    protected void handleCriticalError(String errorMsg, Exception e)
    {
        if (e == null)
        {
            LOGGER.error(errorMsg);
        }
        else
        {
            LOGGER.error(errorMsg, e);
        }
        incrementErrorCounter();
        pipelineStopper.run();
    }


    private void doFirstInternal() throws InterruptedException
    {
        lastDistributedFrameId = getInitialLastDistributedFrameId();
        if (lastDistributedFrameId == Long.MAX_VALUE)
        {
            // Потребители уже имеют всю информацию вплоть до завершения видео
            LOGGER.info("Stopping {} without initialization. All consumers have already received all frames data",
                    distributorName);
            initialLastConsumedFrameIdFuture.complete(Long.MAX_VALUE);
            submitLastFrameId(-1);
            stop();
            return;
        }

        // Определяем идентификатор начального требуемого кадра и количество кадров в обработке

        long initialLastConsumedFrameId = lastDistributedFrameId;
        int processingReqCount = 0;

        KafkaConsumer<?, PerceptionRequest> reqConsumer = KafkaUtils.buildKafkaConsumer(bootstrapServer, username,
                password, clientId, clientId, requestTopicPartition, new PerceptionRequestDeserializer());
        OffsetAndTimestamp initialDistribReqOffset = reqConsumer.offsetsForTimes(Map.of(requestTopicPartition,
                lastDistributedFrameId + 1)).get(requestTopicPartition);

        if (initialDistribReqOffset != null)
        {
            // В топике гарантированно есть записи после последней переданной обработчикам

            long lastWrittenOffset = reqConsumer.endOffsets(Collections.singleton(requestTopicPartition))
                    .get(requestTopicPartition) - 1;

            reqConsumer.seek(requestTopicPartition, initialDistribReqOffset.offset());

            initialFrameIdCycle:
            while (!getDoStop())
            {
                for (ConsumerRecord<?, PerceptionRequest> r : reqConsumer.poll(Duration.ofMillis(100)))
                {
                    if (r.value().getFinished())
                    {
                        initialLastConsumedFrameId = Long.MAX_VALUE;
                        processingReqCount++;
                        break initialFrameIdCycle;
                    }
                    else if (r.value().getFrameId() > initialLastConsumedFrameId)
                    {
                        initialLastConsumedFrameId = r.value().getFrameId();
                        processingReqCount++;
                    }

                    if (lastWrittenOffset == r.offset())
                    {
                        break initialFrameIdCycle;
                    }
                }
            }
        }

        reqConsumer.close();

        if (getDoStop())
        {
            // Завершение initialFrameIdFuture не осуществляем, предполагая, что источники также будут закрываться, и
            // произойдет interrupt.
            return;
        }

        // Может быть равен от -1 до Long.MAX_VALUE
        initialLastConsumedFrameIdFuture.complete(initialLastConsumedFrameId);

        // Инициализируем семафор обрабатываемых запросов, по умолчанию семафор инициализирован в 0
        int permitsCount = maxProcessingRequests - processingReqCount;
        if (permitsCount > 0)
        {
            processingRequestsSem.release(permitsCount);
        }
        else if (permitsCount < 0)
        {
            processingRequestsSem.reducePermits(-permitsCount);
        }

        // Инициализируем поток для отправки кадров на обработку в Kafka
        if (initialLastConsumedFrameId != Long.MAX_VALUE)
        {
            getRequestKafkaSender().start();
        }

        // Инициализируем consumer для чтения результатов обработки кадров

        respConsumer = KafkaUtils.buildKafkaConsumer(bootstrapServer, username,
                password, clientId, clientId, responseTopicPartition, new PerceptionResponseDeserializer());
        if (lastDistributedFrameId == -1)
        {
            respConsumer.seekToBeginning(Collections.singleton(responseTopicPartition));
        }
        else
        {
            // Определение конечного сдвига перед поиском последнего обработанного потребителями кадра минимизирует
            // проблемы при пропаже, по каким-то причинам, из Kafka записей с результатом обработки этого кадра
            long endOffset = respConsumer.endOffsets(Collections.singleton(responseTopicPartition))
                    .get(responseTopicPartition);
            OffsetAndTimestamp lastDistribRespOffsetAndFrameId = respConsumer.offsetsForTimes(
                    Map.of(responseTopicPartition, lastDistributedFrameId)).get(responseTopicPartition);
            if (lastDistribRespOffsetAndFrameId != null)
            {
                respConsumer.seek(responseTopicPartition, lastDistribRespOffsetAndFrameId.offset() +
                        (lastDistribRespOffsetAndFrameId.timestamp() > lastDistributedFrameId ? 0 : 1));
            }
            else
            {
                // Shouldn't happen!
                LOGGER.error("Last distributed result record is missing from the topic [{}]", distributorName);
                respConsumer.seek(responseTopicPartition, endOffset);
            }
        }

        postDoFirst(lastDistributedFrameId);
    }


    @Override
    protected void process()
    {
        for (ConsumerRecord<?, PerceptionResponse> r : respConsumer.poll(Duration.ofMillis(100)))
        {
            if (r.value().getFrameId() != -1 && r.value().hasErr())
            {
                // Пропускаем данный кадр. Потребители должны корректно обработать данную ситуацию
                LOGGER.error("Error occurred during processing [{}]:\n{}", distributorName, r.value().getErr().getMsg());
                incrementErrorCounter();

                if (r.value().getFrameId() <= lastDistributedFrameId)
                {
                    // Данная запись является дубликатом, могла появиться при рестарте сервиса анализа или
                    // перебалансировке партиций на сервисе анализа
                    continue;
                }
            }
            else
            {
                try
                {
                    // Данный метод вызывается даже в случае повторного результата обработки кадра. Предполагается, что
                    // новый результат может отличаться от старого (например, при перезапуске сервиса отслеживания), и
                    // необходимо учесть это изменение в состоянии.
                    modifyDistributorState(r, lastDistributedFrameId);

                    if (getDoStop())
                    {
                        // Необходимо оборвать на данном моменте обработку, поскольку модификация состояния могла
                        // не выполниться полностью
                        return;
                    }

                    if (r.value().getFrameId() != -1 && r.value().getFrameId() <= lastDistributedFrameId)
                    {
                        continue;
                    }

                    sendToConsumers(r.value());
                }
                catch (InterruptedException e)
                {
                    if (!getDoStop())
                    {
                        // Shouldn't happen!
                        handleCriticalError("Interrupted exception occurred but the manager wasn't stopped ["
                                + distributorName + "]. Stopping the hole pipeline.", null);
                    }
                    return;
                }
            }

            processingRequestsSem.release();
            lastDistributedFrameId = r.value().getFrameId();
            submitLastFrameId(r.value().getFrameId());

            if (r.value().getFrameId() == -1)
            {
                stop();
                return;
            }
        }
    }


    private void submitLastFrameId(long frameId)
    {
        metricService.submit(lastFrameIdMetricName, frameId);
    }


    /**
     * Возвращает экземпляр созданного, но не запущенного отправителя запросов на анализ кадров в Kafka.
     * Может быть вызван во время инициализации дистрибьютора.
     *
     * @return отправитель запроса
     */
    protected abstract Actor getRequestKafkaSender();

    /**
     * Опрашивает потребителей и возвращает идентификатор последнего кадра, который был обработан всеми потребителями
     * до последнего выключения.
     *
     * @return идентификатор последнего обработанного кадра или -1, если не был обработан ни один кадр,
     * или {@link Long#MAX_VALUE}, если потребители уже завершили обработку всех кадров видеозаписи
     * @throws InterruptedException в случае прерывания ожидания ответа от потребителей
     */
    protected abstract long getInitialLastDistributedFrameId() throws InterruptedException;

    /**
     * Вызывается непосредственно перед завершением инициализации дистрибьютора.
     *
     * @param lastDistributedFrameId идентификатор последнего переданного обработчикам кадра
     * @throws InterruptedException в случае возникновения прерывания при выполнении инициализации
     */
    protected void postDoFirst(long lastDistributedFrameId) throws InterruptedException {}

    /**
     * Вызывается для обновления внутреннего состояния менеджера в соответствии с новой записью.
     *
     * @param record новая запись
     */
    protected void modifyDistributorState(ConsumerRecord<?, PerceptionResponse> record, long lastDistributedFrameId)
            throws InterruptedException{}

    /**
     * Отправляет результат обработки потребителям.
     *
     * @throws InterruptedException в случае прерывания ожидания принятия результата потребителем
     */
    protected abstract void sendToConsumers(PerceptionResponse response) throws InterruptedException;

    /**
     * Вызывается непосредственно перед выключением дистрибьютора для очистки ресурсов.
     */
    protected abstract void postDoLast();


    @Override
    protected void doLast()
    {
        if (respConsumer != null)
        {
            try
            {
                respConsumer.close();
            }
            catch (Exception e)
            {
                LOGGER.error("Exception occurred during closing consumer [{}]", distributorName, e);
            }
        }

        getRequestKafkaSender().stop();

        postDoLast();
    }


    @Override
    public String getName()
    {
        return distributorName;
    }


    @Override
    protected boolean isDaemon()
    {
        return false;
    }
}