package com.devbeh.managers;


import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.Semaphore;
import java.util.function.BiConsumer;

import com.devbeh.Actor;
import com.devbeh.data.PerceptionRequest;
import com.devbeh.kafka.KafkaUtils;
import com.devbeh.kafka.PerceptionRequestSerializer;
import org.apache.kafka.clients.producer.KafkaProducer;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.common.TopicPartition;


/**
 * Используется для отправки запросов на обработку данных в Kafka в отдельном потоке.
 */
public abstract class AbstractRequestKafkaSender extends Actor
{
    /**
     * Имя менеджера, к которому относится данный отправитель. Используется при логировании и для определения имени
     * потока.
     */
    private final String managerName;

    /**
     * Идентификатор видеозаписи. Используется для определения номера партиции для записи запросов на обработку кадров.
     */
    private final int videoId;

    /**
     * Адрес сервера Kafka.
     */
    private final String bootstrapServer;

    /**
     * Имя пользователя Kafka для SASL аутентификации или пустая строка/null.
     */
    private final String username;

    /**
     * Пароль пользователя Kafka для SASL аутентификации или пустая строка/null.
     */
    private final String password;

    /**
     * Идентификатор клиента Kafka для отправки запросов.
     */
    private final String clientId;

    /**
     * Партиция для записи запросов на обработку кадров.
     */
    private final TopicPartition requestTopicPartition;

    /**
     * Содержит идентификатор первого кадра, который требуется данному компоненту для обработки.
     */
    private final Future<Long> initialLastConsumedFrameIdFuture;

    /**
     * Семафор, используемый для ограничения количества одновременно обрабатываемых запросов.
     */
    private final Semaphore processingRequestsSem;

    /**
     * Обрабатывает критическую ошибку: логирует ошибку, увеличивает счетчик ошибок, останавливает pipeline.
     */
    private final BiConsumer<String, Exception> criticalErrorHandler;

    /**
     * Используется для отправки запросов на обработку.
     */
    private KafkaProducer<?, PerceptionRequest> reqProducer;

    /**
     * Идентификатор последнего отправленного на обработку кадра. Используется для подбора идентификатора
     * завершающего кадра.
     */
    private long lastSentFrameId;


    public AbstractRequestKafkaSender(String managerName, int videoId, String bootstrapServer, String username,
            String password, String clientId, TopicPartition requestTopicPartition,
            Future<Long> initialLastConsumedFrameIdFuture, Semaphore processingRequestsSem,
            BiConsumer<String, Exception> criticalErrorHandler)
    {
        this.managerName = managerName;
        this.videoId = videoId;
        this.bootstrapServer = bootstrapServer;
        this.username = username;
        this.password = password;
        this.clientId = clientId;
        this.requestTopicPartition = requestTopicPartition;
        this.initialLastConsumedFrameIdFuture = initialLastConsumedFrameIdFuture;
        this.processingRequestsSem = processingRequestsSem;
        this.criticalErrorHandler = criticalErrorHandler;
    }


    @Override
    protected void doFirst()
    {
        reqProducer = KafkaUtils.buildKafkaProducer(bootstrapServer, username, password, clientId,
                new PerceptionRequestSerializer());
        try
        {
            lastSentFrameId = initialLastConsumedFrameIdFuture.get();
        }
        catch (Exception e)
        {
            // Shouldn't happen!
            criticalErrorHandler.accept("Unexpected exception occurred during initialization, stopping the hole "
                    + "pipeline [" + managerName + "]", e);
        }
    }


    @Override
    protected void process() throws InterruptedException, ExecutionException
    {
        if (!isAllFrameDataAvailable())
        {
            processingRequestsSem.acquire();
            processingRequestsSem.release();

            ensureNextFrameData();
            if (!processingRequestsSem.tryAcquire())
            {
                // Shouldn't happen!
                criticalErrorHandler.accept("Unexpected situation: permit is not available [" + managerName + "]",
                        null);
                return;
            }
        }

        // В данной реализации videoId используется также в качестве managerId
        PerceptionRequest.Builder rb = PerceptionRequest.newBuilder().setManagerId(videoId).setVideoId(videoId);
        long frameId;
        if (getCurrentFrameId() == -1)
        {
            frameId = lastSentFrameId + 1;
            rb.setFinished(true);
        }
        else
        {
            frameId = getCurrentFrameId();
            lastSentFrameId = frameId;
            buildRequestPart(rb);
        }
        rb.setFrameId(frameId);

        // Цикл вызова process может застрять на переотправке данной записи, если ее размер превышает максимум.
        // Считаем, что все отправляемые сообщения будут укладываться в ограничение.
        reqProducer.send(new ProducerRecord<>(requestTopicPartition.topic(), requestTopicPartition.partition(),
                frameId, null, rb.build())).get();

        // Очищаем подготовленные данные кадра, тем самым переходя на обработку следующего кадра
        clearFrameData();
    }


    /**
     * Определяет, сохранены ли все данные следующего обрабатываемого кадра.
     *
     * @return true - данные сохранены, false - необходимо вызвать {@link #ensureNextFrameData()}
     */
    protected abstract boolean isAllFrameDataAvailable();

    /**
     * Забирает данные следующего обрабатываемого кадра из очередей. В случае возникновения исключения, после
     * которого {@link #isAllFrameDataAvailable} вернет false, данный метод будет вызван повторно для завершения
     * инициализации.
     *
     * @throws InterruptedException в случае прерывания получения данных кадра из очереди
     */
    protected abstract void ensureNextFrameData() throws InterruptedException;

    /**
     * Возвращает идентификатор кадра, к которому относятся сохраненные данные.
     *
     * @return идентификатор текущего кадра
     */
    protected abstract long getCurrentFrameId();

    /**
     * Очищает текущие сохраненные данные кадра.
     */
    protected abstract void clearFrameData();

    /**
     * Записывает в переданный билдер запрос в соответствии с текущими инициализированными данными.
     * Гарантируется, что данный метод будет вызван только при сохраненных данных кадра, который не является
     * завершающим (идентификатор отличен от -1).
     *
     * @param rb билдер
     */
    protected abstract void buildRequestPart(PerceptionRequest.Builder rb);

    /**
     * Вызывается непосредственно перед завершением работы отправителя для очистки ресурсов.
     */
    protected void postDoLast() {}


    @Override
    protected void doLast()
    {
        try
        {
            reqProducer.close();
        }
        finally
        {
            postDoLast();
        }
    }


    @Override
    public String getName()
    {
        return managerName;
    }


    @Override
    protected boolean isDaemon()
    {
        return false;
    }
}
