package com.devbeh.managers;


import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Future;
import java.util.concurrent.Semaphore;

import com.devbeh.Actor;
import com.devbeh.DetectionClass;
import com.devbeh.MetricService;
import com.devbeh.ReducibleSemaphore;
import com.devbeh.data.DistanceEstimationEntry;
import com.devbeh.data.PerceptionRequest;
import com.devbeh.data.PerceptionResponse;
import com.devbeh.data.PerceptionResponseOrBuilder;
import com.devbeh.data.TrackingEntry;
import org.apache.kafka.common.TopicPartition;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


public class ActionsManager extends AbstractDistributor implements TrackingConsumer, DistanceDataConsumer
{
    private static final Logger LOGGER = LoggerFactory.getLogger(ActionsManager.class);

    // Получатели результатов детектирования
    private final KBManager kbm;
    private final PerceptionVisualizationManager pvm;

    /**
     * Очередь результатов отслеживания.
     */
    private final BlockingQueue<PerceptionResponseOrBuilder> trackingQueue;

    /**
     * Очередь результатов оценки трехмерных координат ключевых точек объектов в пространстве камеры.
     */
    private final BlockingQueue<PerceptionResponse> distancesQueue;

    /**
     * Используется для отправки данных для анализа компоненту детектирования.
     */
    private final ActionsRequestKafkaSender requestKafkaSender;


    public ActionsManager(String videoName, int videoId, String bootstrapServer, String username, String password,
            String requestTopicName, String responseTopicPrefix, int maxQueuedRequests, int maxProcessingRequests,
            KBManager kbm, PerceptionVisualizationManager pvm, Runnable pipelineStopper, MetricService metricService)
    {
        this(videoName, videoId, bootstrapServer, username, password, "actions-manager-" + videoId,
                new TopicPartition(requestTopicName, videoId), new TopicPartition(responseTopicPrefix + videoId, 0),
                new CompletableFuture<>(), maxQueuedRequests, maxProcessingRequests, new ReducibleSemaphore(0),
                kbm, pvm, pipelineStopper, metricService);
    }

    private ActionsManager(String videoName, int videoId, String bootstrapServer, String username, String password,
            String clientId, TopicPartition requestTopicPartition, TopicPartition responseTopicPartition,
            CompletableFuture<Long> initialLastConsumedFrameIdFuture, int maxQueuedRequests, int maxProcessingRequests,
            ReducibleSemaphore processingRequestsSem, KBManager kbm, PerceptionVisualizationManager pvm,
            Runnable pipelineStopper, MetricService metricService)
    {
        super("ActionsDistributor-" + videoName, videoName + ".actions.",
                bootstrapServer, username, password, clientId, requestTopicPartition, responseTopicPartition,
                initialLastConsumedFrameIdFuture, maxProcessingRequests, processingRequestsSem, pipelineStopper,
                metricService);

        this.kbm = kbm;
        this.pvm = pvm;
        this.trackingQueue = new ArrayBlockingQueue<>(maxQueuedRequests - 1);
        this.distancesQueue = new ArrayBlockingQueue<>(maxQueuedRequests - 1);
        this.requestKafkaSender = new ActionsRequestKafkaSender("ActionsManager-" + videoName, videoId,
                bootstrapServer, username, password, clientId, requestTopicPartition, initialLastConsumedFrameIdFuture,
                processingRequestsSem);
    }


    @Override
    public void putTracking(PerceptionResponseOrBuilder tracking) throws InterruptedException
    {
        if (!getDoStop())
        {
            // Если данный менеджер завершил свою работу, игнорируем дальнейшие сообщения
            trackingQueue.put(tracking);
        }
    }


    @Override
    public void putDistances(PerceptionResponse distanceData) throws InterruptedException
    {
        if (!getDoStop())
        {
            // Если данный менеджер завершил свою работу, игнорируем дальнейшие сообщения
            distancesQueue.put(distanceData);
        }
    }


    @Override
    protected Actor getRequestKafkaSender()
    {
        return requestKafkaSender;
    }


    @Override
    protected long getInitialLastDistributedFrameId() throws InterruptedException
    {
        return Math.min(kbm.getInitialLastConsumedFrameId(), pvm.getInitialLastConsumedFrameId());
    }


    @Override
    protected void sendToConsumers(PerceptionResponse response) throws InterruptedException
    {
        sendActionsIfRequired(kbm, response);
        sendActionsIfRequired(pvm, response);
    }


    public static void sendActionsIfRequired(ActionsConsumer actionsConsumer, PerceptionResponseOrBuilder actions)
            throws InterruptedException
    {
        if (actions.getFrameId() == -1 || actionsConsumer.getInitialLastConsumedFrameId() < actions.getFrameId())
        {
            actionsConsumer.putActions(actions);
        }
    }


    @Override
    protected void postDoLast()
    {
        // Очищаем для того, чтобы менеджер-источник не сидел на блокировке в случае переотправки завершающего сообщения
        distancesQueue.clear();
    }


    /**
     * Используется для отправки данных для анализа компоненту детектирования.
     */
    private class ActionsRequestKafkaSender extends AbstractRequestKafkaSender
    {
        /**
         * Имя менеджера, используется при логировании.
         */
        private final String managerName;

        /**
         * Результат отслеживания для текущего обрабатываемого кадра или null.
         */
        private PerceptionResponseOrBuilder currentTrackingResult;

        /**
         * Результаты оценки расстояния для текущего кадра или null.
         */
        private PerceptionResponse currentDistanceResult;


        public ActionsRequestKafkaSender(String managerName, int videoId, String bootstrapServer, String username,
                String password, String clientId, TopicPartition requestTopicPartition,
                Future<Long> initialLastConsumedFrameIdFuture, Semaphore processingRequestsSem)
        {
            super(managerName, videoId, bootstrapServer, username, password, clientId, requestTopicPartition,
                    initialLastConsumedFrameIdFuture, processingRequestsSem, ActionsManager.this::handleCriticalError);
            this.managerName = managerName;
        }


        @Override
        protected boolean isAllFrameDataAvailable()
        {
            return currentTrackingResult != null && currentDistanceResult != null
                    && currentTrackingResult.getFrameId() == currentDistanceResult.getFrameId();
        }


        @SuppressWarnings("DuplicatedCode")
        @Override
        protected void ensureNextFrameData() throws InterruptedException
        {
            while (!isAllFrameDataAvailable())
            {
                if (currentTrackingResult == null)
                {
                    currentTrackingResult = trackingQueue.take();
                }

                if (currentDistanceResult == null)
                {
                    currentDistanceResult = distancesQueue.take();
                }

                if (currentDistanceResult.getFrameId() != currentTrackingResult.getFrameId())
                {
                    if (currentDistanceResult.getFrameId() == -1 || currentTrackingResult.getFrameId() != -1
                            && currentTrackingResult.getFrameId() < currentDistanceResult.getFrameId())
                    {
                        LOGGER.warn("Missing a distance result with id {} [{}]", currentTrackingResult.getFrameId(),
                                managerName);
                        currentTrackingResult = trackingQueue.take();
                    }
                    else
                    {
                        LOGGER.warn("Missing a tracking with frame id {} [{}]", currentDistanceResult.getFrameId(),
                                managerName);
                        currentDistanceResult = distancesQueue.take();
                    }
                }
            }
        }


        @Override
        protected long getCurrentFrameId()
        {
            return currentDistanceResult.getFrameId();
        }


        @Override
        protected void clearFrameData()
        {
            currentTrackingResult = null;
            currentDistanceResult = null;
        }


        @Override
        protected void buildRequestPart(PerceptionRequest.Builder rb)
        {
            for (int i = 0; i < currentTrackingResult.getTracking().getEntriesCount(); i++)
            {
                TrackingEntry te = currentTrackingResult.getTracking().getEntries(i);
                DistanceEstimationEntry de = currentDistanceResult.getDistance().getEntries(i);

                if (te.getDetection().getClassId() == DetectionClass.PERSON.getId())
                {
                    rb.getActionsBuilder().addEntriesBuilder().setId(te.getId()).addAllJoints(de.getPointsList());
                }
            }
        }
    }
}
