package com.devbeh.managers;


import java.util.List;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Future;
import java.util.concurrent.Semaphore;

import com.devbeh.Actor;
import com.devbeh.DetectionClass;
import com.devbeh.FrameIdAndBytes;
import com.devbeh.MetricService;
import com.devbeh.ReducibleSemaphore;
import com.devbeh.data.DetectionEntry;
import com.devbeh.data.PerceptionRequest;
import com.devbeh.data.PerceptionResponse;
import com.devbeh.data.PerceptionResponseOrBuilder;
import com.google.protobuf.ByteString;
import org.apache.kafka.common.TopicPartition;


/**
 * Отправляет кадры на обработку сервису детектирования людей и объектов, ожидает результат обработки и пересылает его
 * остальным менеджерам.
 */
public class DetectionManager extends AbstractDistributor implements FrameConsumer
{
    // Получатели результатов детектирования
    private final TrackingManager tm;
    private final PoseEstimationManager pem;

    /**
     * Очередь кадров для обработки.
     */
    private final BlockingQueue<FrameIdAndBytes> frameQueue;

    /**
     * Используется для отправки данных для анализа компоненту детектирования.
     */
    private final DetectionRequestKafkaSender requestKafkaSender;


    public DetectionManager(String videoName, int videoId, String bootstrapServer, String username, String password,
            String requestTopicName, String responseTopicPrefix, int maxQueuedRequests, int maxProcessingRequests,
            TrackingManager tm, PoseEstimationManager pem, Runnable pipelineStopper, MetricService metricService)
    {
        this(videoName, videoId, bootstrapServer, username, password, "detection-manager-" + videoId,
                new TopicPartition(requestTopicName, videoId), new TopicPartition(responseTopicPrefix + videoId, 0),
                new CompletableFuture<>(), maxQueuedRequests, maxProcessingRequests, new ReducibleSemaphore(0),
                tm, pem, pipelineStopper, metricService);
    }

    private DetectionManager(String videoName, int videoId, String bootstrapServer, String username, String password,
            String clientId, TopicPartition requestTopicPartition, TopicPartition responseTopicPartition,
            CompletableFuture<Long> initialLastConsumedFrameIdFuture, int maxQueuedRequests, int maxProcessingRequests,
            ReducibleSemaphore processingRequestsSem, TrackingManager tm, PoseEstimationManager pem,
            Runnable pipelineStopper, MetricService metricService)
    {
        super("DetectionDistributor-" + videoName, videoName + ".detection.",
                bootstrapServer, username, password, clientId, requestTopicPartition, responseTopicPartition,
                initialLastConsumedFrameIdFuture, maxProcessingRequests, processingRequestsSem, pipelineStopper,
                metricService);

        this.tm = tm;
        this.pem = pem;
        this.frameQueue = new ArrayBlockingQueue<>(maxQueuedRequests - 1);
        this.requestKafkaSender = new DetectionRequestKafkaSender("DetectionManager-" + videoName, videoId,
                bootstrapServer, username, password, clientId, requestTopicPartition, initialLastConsumedFrameIdFuture,
                processingRequestsSem);
    }


    @Override
    public void putFrame(FrameIdAndBytes frameIdAndBytes) throws InterruptedException
    {
        if (!getDoStop())
        {
            // Если данный менеджер завершил свою работу, игнорируем дальнейшие сообщения
            frameQueue.put(frameIdAndBytes);
        }
    }


    @Override
    protected Actor getRequestKafkaSender()
    {
        return requestKafkaSender;
    }


    @Override
    protected long getInitialLastDistributedFrameId() throws InterruptedException
    {
        return Math.min(tm.getInitialLastConsumedFrameId(), pem.getInitialLastConsumedFrameId());
    }


    @Override
    protected void sendToConsumers(PerceptionResponse response) throws InterruptedException
    {
        // Удаляем объекты ненужных типов
        PerceptionResponseOrBuilder correctedResponse = response;
        if (response.getFrameId() != -1)
        {
            List<DetectionEntry> allDetections = response.getDetection().getEntriesList();

            PerceptionResponse.Builder responseBuilder = response.toBuilder();
            responseBuilder.getDetectionBuilder().clearEntries();

            for (DetectionEntry de : allDetections)
            {
                if (DetectionClass.findById(de.getClassId()) != null)
                {
                    responseBuilder.getDetectionBuilder().addEntries(de);
                }
            }
            correctedResponse = responseBuilder;
        }

        sendDetectionIfRequired(tm, correctedResponse);
        sendDetectionIfRequired(pem, correctedResponse);
    }


    public static void sendDetectionIfRequired(DetectionConsumer detectionConsumer,
            PerceptionResponseOrBuilder detection)
            throws InterruptedException
    {
        if (detection.getFrameId() == -1 || detectionConsumer.getInitialLastConsumedFrameId() < detection.getFrameId())
        {
            detectionConsumer.putDetection(detection);
        }
    }


    @Override
    protected void postDoLast()
    {
        // Очищаем для того, чтобы менеджер-источник не застрял на блокировке в случае переотправки завершающего
        // сообщения
        frameQueue.clear();
    }


    /**
     * Используется для отправки данных для анализа компоненту детектирования.
     */
    private class DetectionRequestKafkaSender extends AbstractRequestKafkaSender
    {
        /**
         * Текущий обрабатываемый кадр или null.
         */
        private FrameIdAndBytes currentFrameInfo;


        public DetectionRequestKafkaSender(String managerName, int videoId, String bootstrapServer, String username,
                String password, String clientId, TopicPartition requestTopicPartition,
                Future<Long> initialLastConsumedFrameIdFuture, Semaphore processingRequestsSem)
        {
            super(managerName, videoId, bootstrapServer, username, password, clientId, requestTopicPartition,
                    initialLastConsumedFrameIdFuture, processingRequestsSem,
                    DetectionManager.this::handleCriticalError);
        }


        @Override
        protected boolean isAllFrameDataAvailable()
        {
            return currentFrameInfo != null;
        }


        @Override
        protected void ensureNextFrameData() throws InterruptedException
        {
            currentFrameInfo = frameQueue.take();
        }


        @Override
        protected long getCurrentFrameId()
        {
            return currentFrameInfo.frameId;
        }


        @Override
        protected void clearFrameData()
        {
            currentFrameInfo = null;
        }


        @Override
        protected void buildRequestPart(PerceptionRequest.Builder rb)
        {
            rb.getDetectionBuilder().setImage(ByteString.copyFrom(currentFrameInfo.frame));
        }
    }
}
