package com.devbeh.managers;


import java.io.File;
import java.time.Duration;
import java.util.Collections;
import java.util.Iterator;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.atomic.AtomicLong;

import com.devbeh.Actor;
import com.devbeh.FrameIdAndBytes;
import com.devbeh.MetricService;
import com.devbeh.commondata.DevBehDetections;
import com.devbeh.commondata.DevBehEntry;
import com.devbeh.commondata.DevBehFrameInfo;
import com.devbeh.commondata.DevBehFrameObject;
import com.devbeh.kafka.DevBehDetectionsDeserializer;
import com.devbeh.kafka.KafkaUtils;
import com.devbeh.videoproviders.VideoProvider;
import org.apache.kafka.clients.consumer.ConsumerRecord;
import org.apache.kafka.clients.consumer.KafkaConsumer;
import org.apache.kafka.clients.consumer.OffsetAndTimestamp;
import org.apache.kafka.common.TopicPartition;
import org.bytedeco.javacv.FFmpegFrameRecorder;
import org.bytedeco.javacv.Frame;
import org.bytedeco.javacv.OpenCVFrameConverter;
import org.bytedeco.opencv.global.opencv_imgproc;
import org.bytedeco.opencv.opencv_core.Point;
import org.bytedeco.opencv.opencv_core.Rect;
import org.bytedeco.opencv.opencv_core.Scalar;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import static com.devbeh.managers.PerceptionVisualizationManager.COLORS;


/**
 * Считывает результаты детектирования девиантного поведения людей, для каждого детектированного случая сохраняет
 * отрывок видеозаписи с отмеченными обрамляющими рамками в заданной директории.
 */
public class DetectionsVisualizationManager extends Actor
{
    private static final Logger LOGGER = LoggerFactory.getLogger(DetectionsVisualizationManager.class);

    /**
     * Осуществляет преобразование кадра из Mat формата.
     */
    private final OpenCVFrameConverter.ToMat matConverter = new OpenCVFrameConverter.ToMat();

    /**
     * Счетчик ошибок.
     */
    private final AtomicLong errorCounter = new AtomicLong();

    /**
     * Имя данного менеджера.
     */
    private final String managerName;

    /**
     * Адрес сервера Kafka.
     */
    private final String bootstrapServer;

    /**
     * Имя пользователя Kafka для SASL аутентификации или пустая строка/null.
     */
    private final String username;

    /**
     * Пароль пользователя Kafka для SASL аутентификации или пустая строка/null.
     */
    private final String password;

    /**
     * Идентификатор клиента (и группы) Kafka, используемый для работы с топиками состояния и результатов
     * детектирования.
     */
    private final String clientId;

    /**
     * Топик для хранения результатов детектирования девиантного поведения людей.
     */
    private final TopicPartition devBehDetectionsTopicPartition;

    /**
     * Объект директории, в которой необходимо разместить отрывки видеозаписей с детектированными случаями девиантного
     * поведения.
     */
    private final File detectionsDirFile;

    /**
     * Номер кадра, с которого необходимо обрабатывать детектированные случаи девиантного поведения.
     */
    private final long startFrameId;

    /**
     * Используется для получения кадров видеозаписи.
     */
    private final VideoProvider videoProvider;

    /**
     * Передается экземпляру {@link VideoProvider} для инициализации.
     */
    private final int maxFrameSizePixels;

    /**
     * Используется для остановки конвейера обработки видео в случае критичной ошибки.
     */
    private final Runnable pipelineStopper;

    /**
     * Используется для публикации метрик.
     */
    private final MetricService metricService;

    /**
     * Имя метрики последнего обработанного идентификатора кадра.
     */
    private final String lastFrameIdMetricName;

    /**
     * Имя метрики количества ошибок.
     */
    private final String totalErrorsMetricName;

    /**
     * Используется для чтения детектированных случаев девиантного поведения.
     */
    private KafkaConsumer<?, DevBehDetections> devBehDetectionConsumer;

    /**
     * Идентификатор последнего кадра, для которого было получено и обработано сообщение с результатами детектирования.
     */
    private long currentFrameId;


    public DetectionsVisualizationManager(String videoName, int videoId, String bootstrapServer, String username,
            String password, String devBehDetectionsTopicPrefix, String detectionsDirPath, long startFrameId,
            VideoProvider videoProvider, int maxFrameSizePixels, Runnable pipelineStopper, MetricService metricService)
    {
        this.managerName = "DetectionsVisualManager-" + videoName;
        this.bootstrapServer = bootstrapServer;
        this.username = username;
        this.password = password;
        this.clientId = "dvm-" + videoId;
        this.devBehDetectionsTopicPartition = new TopicPartition(devBehDetectionsTopicPrefix + videoId, 0);
        this.detectionsDirFile = new File(detectionsDirPath, videoName);
        this.startFrameId = startFrameId;
        this.videoProvider = videoProvider;
        this.maxFrameSizePixels = maxFrameSizePixels;
        this.pipelineStopper = pipelineStopper;
        this.metricService = metricService;
        String metricsPrefix = videoName + ".dvm.";
        this.lastFrameIdMetricName = metricsPrefix + "lastFrameId";
        this.totalErrorsMetricName = metricsPrefix + "totalErrors";
        metricService.submit(totalErrorsMetricName, 0);
    }


    @Override
    protected void doFirst()
    {
        if (startFrameId < 0 || startFrameId == Long.MAX_VALUE)
        {
            return;
        }

        try
        {
            // Создаем необходимые директории
            if (!detectionsDirFile.exists())
            {
                if (!detectionsDirFile.mkdirs())
                {
                    LOGGER.error("Can't create directory '{}' [{}]", detectionsDirFile.getAbsolutePath(), managerName);
                    incrementErrorCounter();
                    pipelineStopper.run();
                    return;
                }
            }

            // Инициализируем читателя топика результатов детектирования
            currentFrameId = startFrameId - 1;
            devBehDetectionConsumer = KafkaUtils.buildKafkaConsumer(bootstrapServer, username,
                    password, clientId, clientId, devBehDetectionsTopicPartition, new DevBehDetectionsDeserializer());
            Long endOffset = devBehDetectionConsumer.endOffsets(Collections.singleton(devBehDetectionsTopicPartition))
                    .get(devBehDetectionsTopicPartition);
            OffsetAndTimestamp startOffset = devBehDetectionConsumer.offsetsForTimes(
                    Map.of(devBehDetectionsTopicPartition, startFrameId)).get(devBehDetectionsTopicPartition);
            devBehDetectionConsumer.seek(devBehDetectionsTopicPartition, startOffset == null ? endOffset
                    : startOffset.offset());
        }
        catch (Exception e)
        {
            LOGGER.error("Exception occurred during initializing manager [{}]", managerName, e);
            incrementErrorCounter();
            pipelineStopper.run();
        }
    }


    /**
     * Увеличивает счетчик ошибок и обновляет соответствущую метрику.
     */
    private void incrementErrorCounter()
    {
        metricService.submit(totalErrorsMetricName, errorCounter.incrementAndGet());
    }


    /**
     * Отображает в метриках последний обработанный кадр видеозаписи.
     */
    private void submitLastFrameId(long frameId)
    {
        metricService.submit(lastFrameIdMetricName, frameId);
    }


    @Override
    protected void process()
    {
        for (ConsumerRecord<?, DevBehDetections> r : devBehDetectionConsumer.poll(Duration.ofMillis(100)))
        {
            if (currentFrameId >= r.timestamp())
            {
                continue;
            }
            currentFrameId = r.timestamp();

            for (DevBehEntry entry : r.value().getEntriesList())
            {
                try
                {
                    processDetectionEntry(entry);
                }
                catch (Exception e)
                {
                    if (!getDoStop())
                    {
                        LOGGER.error("Exception occurred during detection visualization. Stopping the hole pipeline. "
                                        + "[{}]", managerName, e);
                        incrementErrorCounter();
                        pipelineStopper.run();
                    }
                    return;
                }
            }

            if (r.value().getFrameId() == -1)
            {
                currentFrameId = Long.MAX_VALUE;
                submitLastFrameId(-1);
                stop();
                break;
            }
            else
            {
                submitLastFrameId(currentFrameId);
            }
        }
    }


    /**
     * Создает объекты для чтения кадров и записи видео и вызывает обработчик
     * {@link #processDetection(VideoProvider, FFmpegFrameRecorder, DevBehEntry)}.
     *
     * @param entry детектированный инцидент девиантного поведения
     */
    private void processDetectionEntry(DevBehEntry entry) throws Exception
    {
        File resultVideoFile = new File(detectionsDirFile, entry.getType() + "_" + entry.getStartFrameId()
                + "-" + entry.getEndFrameId() + ".mp4");
        if (resultVideoFile.exists())
        {
            if (!resultVideoFile.delete())
            {
                LOGGER.error("Can't delete existing detection file '{}', skip detection. [{}]",
                        resultVideoFile.getAbsolutePath(), managerName);
                incrementErrorCounter();
                return;
            }
        }

        videoProvider.initVideoInfo(maxFrameSizePixels);
        videoProvider.initInitialFrameId(entry.getStartFrameId());
        try
        {
            FFmpegFrameRecorder frameRecorder = new FFmpegFrameRecorder(resultVideoFile,
                    videoProvider.getWidth(), videoProvider.getHeight());
            try (frameRecorder)
            {
                frameRecorder.setFormat("mp4");
                frameRecorder.setFrameRate(videoProvider.getFps());
                frameRecorder.setVideoQuality(0);
                frameRecorder.start();

                processDetection(videoProvider, frameRecorder, entry);
            }
        }
        finally
        {
            videoProvider.close();
        }
    }


    /**
     * Записывает кадры с отмеченными действующими лицами и объектами в выходной видеопоток детектированного инцидента.
     *
     * @param vp источник кадров видеозаписи
     * @param frameRecorder выходной видеопоток
     * @param devBehEntry информация о детектированном инциденте
     */
    private void processDetection(VideoProvider vp, FFmpegFrameRecorder frameRecorder, DevBehEntry devBehEntry)
            throws Exception
    {
        Iterator<DevBehFrameInfo> frameInfoIterator = devBehEntry.getFrameInfoList().listIterator();
        DevBehFrameInfo currentFrameInfo = null;
        while (!getDoStop())
        {
            Optional<FrameIdAndBytes> frameData = vp.read(100);
            if (frameData.isEmpty())
            {
                continue;
            }

            if (frameData.get().frameId > devBehEntry.getEndFrameId() || frameData.get().frameId == -1)
            {
                // Интервал девиантного поведения завершился
                if (frameData.get().frameMat != null)
                {
                    frameData.get().frameMat.close();
                }
                return;
            }

            while (currentFrameInfo == null && frameInfoIterator.hasNext())
            {
                currentFrameInfo = frameInfoIterator.next();
                if (currentFrameInfo.getFrameId() < frameData.get().frameId)
                {
                    currentFrameInfo = null;
                }
            }

            if (currentFrameInfo != null && currentFrameInfo.getFrameId() == frameData.get().frameId)
            {
                for (DevBehFrameObject obj : currentFrameInfo.getObjectsList())
                {
                    int tlX = (int) obj.getBox().getTopLeft().getX();
                    int tlY = (int) obj.getBox().getTopLeft().getY();
                    int brX = (int) obj.getBox().getBottomRight().getX();
                    int brY = (int) obj.getBox().getBottomRight().getY();

                    Scalar color = COLORS[(int) (obj.getId() % COLORS.length)];
                    opencv_imgproc.rectangle(frameData.get().frameMat,
                            new Rect(tlX, tlY, brX - tlX, brY - tlY),
                            color, 1, opencv_imgproc.LINE_8, 0);
                    opencv_imgproc.putText(frameData.get().frameMat, Long.toString(obj.getId()),
                            new Point(tlX, Math.max(15, tlY)), opencv_imgproc.FONT_HERSHEY_SIMPLEX, 0.4, color);
                }
                currentFrameInfo = null;
            }

            Frame f = matConverter.convert(frameData.get().frameMat);
            frameRecorder.record(f);
            frameData.get().frameMat.close();
        }

        if (getDoStop())
        {
            // Останавливаем обработку
            throw new RuntimeException();
        }
    }


    @Override
    public void doLast()
    {
        matConverter.close();
        if (devBehDetectionConsumer != null)
        {
            devBehDetectionConsumer.close();
        }
    }


    @Override
    public String getName()
    {
        return managerName;
    }


    @Override
    protected boolean isDaemon()
    {
        return false;
    }
}
