package com.devbeh.managers;


import java.util.Map;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Future;
import java.util.concurrent.Semaphore;
import java.util.stream.Collectors;

import com.devbeh.Actor;
import com.devbeh.DbmProperties;
import com.devbeh.DetectionClass;
import com.devbeh.FrameIdAndBytes;
import com.devbeh.MetricService;
import com.devbeh.ReducibleSemaphore;
import com.devbeh.data.DistanceEstimationEntry;
import com.devbeh.data.PerceptionRequest;
import com.devbeh.data.PerceptionResponse;
import com.devbeh.data.PerceptionResponseOrBuilder;
import com.devbeh.data.PoseEstimationEntry;
import com.devbeh.data.TrackingEntryOrBuilder;
import com.google.protobuf.ByteString;
import org.apache.kafka.common.TopicPartition;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


public class DistanceEstimationManager extends AbstractDistributor implements FrameConsumer, TrackingConsumer,
        PoseDataConsumer
{
    private static final Logger LOGGER = LoggerFactory.getLogger(DistanceEstimationManager.class);

    // Получатели результатов оценки поз
    private final ActionsManager am;
    private final PerceptionVisualizationManager pvm;
    private final KBManager kbm;

    /**
     * Очередь кадров для обработки.
     */
    private final BlockingQueue<FrameIdAndBytes> frameQueue;

    /**
     * Очередь результатов отслеживания.
     */
    private final BlockingQueue<PerceptionResponseOrBuilder> trackingQueue;

    /**
     * Очередь результатов оценки скелетов людей.
     */
    private final BlockingQueue<PerceptionResponse> poseQueue;

    /**
     * Используется для отправки данных для анализа компоненту оценки расстояния.
     */
    private final DistanceEstimationRequestKafkaSender requestKafkaSender;

    /**
     * Высота размещения камеры над плоскостью земли в миллиметрах.
     */
    private final float cameraHeightMm;

    /**
     * Фокусное расстояние камеры в миллиметрах.
     */
    private final float focalLengthMm;

    /**
     * Размер пикселей в миллиметрах.
     */
    private final float pixelSizeMm;

    /**
     * Вектор нормали поверхности земли в пространстве камеры или null.
     */
    private final DbmProperties.NormalVector normal;

    /**
     * Используется для вычисления точки касания объекта с землей. Считается, что точка касания расположена посередине
     * обрамляющей рамки по оси x и на данный процент выше нижней точки рамки. Процент вычисляется относительно высоты
     * рамки.
     */
    private final float boxContactPointLiftPercent;


    public DistanceEstimationManager(String videoName, int videoId, float cameraHeightMm, float focalLengthMm,
            float pixelSizeMm, DbmProperties.NormalVector normal, String bootstrapServer,
            String username, String password, String requestTopicName, String responseTopicPrefix,
            float boxContactPointLiftPercent, int maxQueuedRequests, int maxProcessingRequests, ActionsManager am,
            KBManager kbm, PerceptionVisualizationManager pvm, Runnable pipelineStopper, MetricService metricService)
    {
        this(videoName, videoId, cameraHeightMm, focalLengthMm, pixelSizeMm, normal, bootstrapServer, username,
                password, "distance-manager-" + videoId, new TopicPartition(requestTopicName, videoId),
                new TopicPartition(responseTopicPrefix + videoId, 0), boxContactPointLiftPercent,
                new CompletableFuture<>(), maxQueuedRequests, maxProcessingRequests, new ReducibleSemaphore(0),
                am, kbm, pvm, pipelineStopper, metricService);
    }

    private DistanceEstimationManager(String videoName, int videoId, float cameraHeightMm, float focalLengthMm,
            float pixelSizeMm, DbmProperties.NormalVector normal, String bootstrapServer, String username,
            String password, String clientId, TopicPartition requestTopicPartition,
            TopicPartition responseTopicPartition, float boxContactPointLiftPercent,
            CompletableFuture<Long> initialFrameIdFuture, int maxQueuedRequests, int maxProcessingRequests,
            ReducibleSemaphore processingRequestsSem, ActionsManager am, KBManager kbm,
            PerceptionVisualizationManager pvm, Runnable pipelineStopper, MetricService metricService)
    {
        super("DistanceEstimationDistributor-" + videoName, videoName + ".distance.",
                bootstrapServer, username, password, clientId, requestTopicPartition, responseTopicPartition,
                initialFrameIdFuture, maxProcessingRequests, processingRequestsSem, pipelineStopper, metricService);

        this.am = am;
        this.kbm = kbm;
        this.pvm = pvm;
        this.frameQueue = new ArrayBlockingQueue<>(maxQueuedRequests - 1);
        this.trackingQueue = new ArrayBlockingQueue<>(maxQueuedRequests - 1);
        this.poseQueue = new ArrayBlockingQueue<>(maxQueuedRequests - 1);
        this.requestKafkaSender = new DistanceEstimationRequestKafkaSender("DistanceEstimationManager-" + videoName,
                videoId, bootstrapServer, username, password, clientId, requestTopicPartition, initialFrameIdFuture,
                processingRequestsSem);
        this.cameraHeightMm = cameraHeightMm;
        this.focalLengthMm = focalLengthMm;
        this.pixelSizeMm = pixelSizeMm;
        this.normal = normal;
        this.boxContactPointLiftPercent = boxContactPointLiftPercent;
    }


    @Override
    public void putFrame(FrameIdAndBytes frameIdAndBytes) throws InterruptedException
    {
        if (!getDoStop())
        {
            // Если данный менеджер завершил свою работу, игнорируем дальнейшие сообщения
            frameQueue.put(frameIdAndBytes);
        }
    }


    @Override
    public void putTracking(PerceptionResponseOrBuilder tracking) throws InterruptedException
    {
        if (!getDoStop())
        {
            // Если данный менеджер завершил свою работу, игнорируем дальнейшие сообщения
            trackingQueue.put(tracking);
        }
    }


    @Override
    public void putPoses(PerceptionResponse poses) throws InterruptedException
    {
        if (!getDoStop())
        {
            // Если данный менеджер завершил свою работу, игнорируем дальнейшие сообщения
            poseQueue.put(poses);
        }
    }


    @Override
    protected Actor getRequestKafkaSender()
    {
        return requestKafkaSender;
    }


    @Override
    protected long getInitialLastDistributedFrameId() throws InterruptedException
    {
        return Math.min(
                Math.min(am.getInitialLastConsumedFrameId(), kbm.getInitialLastConsumedFrameId()),
                pvm.getInitialLastConsumedFrameId());
    }


    @Override
    protected void sendToConsumers(PerceptionResponse response) throws InterruptedException
    {
        sendDistancesIfRequired(am, response);
        sendDistancesIfRequired(kbm, response);
        sendDistancesIfRequired(pvm, response);
    }


    public static void sendDistancesIfRequired(DistanceDataConsumer distanceDataConsumer, PerceptionResponse distances)
            throws InterruptedException
    {
        if (distances.getFrameId() == -1 || distanceDataConsumer.getInitialLastConsumedFrameId()
                < distances.getFrameId())
        {
            distanceDataConsumer.putDistances(distances);
        }
    }


    @Override
    protected void postDoLast()
    {
        // Очищаем для того, чтобы менеджер-источник не сидел на блокировке в случае переотправки завершающего сообщения
        frameQueue.clear();
        trackingQueue.clear();
        poseQueue.clear();
    }


    /**
     * Используется для отправки данных для анализа компоненту детектирования.
     */
    private class DistanceEstimationRequestKafkaSender extends AbstractRequestKafkaSender
    {
        /**
         * Имя менеджера, используется при логировании.
         */
        private final String managerName;

        /**
         * Текущий обрабатываемый кадр или null.
         */
        private FrameIdAndBytes currentFrameInfo;

        /**
         * Результат отслеживания для текущего обрабатываемого кадра или null.
         */
        private PerceptionResponseOrBuilder currentTrackingResult;

        /**
         * Результат оценки трехмерных поз людей для текущего обрабатываемого кадра или null.
         */
        private PerceptionResponse currentPoseData;


        public DistanceEstimationRequestKafkaSender(String managerName, int videoId, String bootstrapServer,
                String username, String password, String clientId, TopicPartition requestTopicPartition,
                Future<Long> initialFrameIdFuture, Semaphore processingRequestsSem)
        {
            super(managerName, videoId, bootstrapServer, username, password, clientId, requestTopicPartition,
                    initialFrameIdFuture, processingRequestsSem, DistanceEstimationManager.this::handleCriticalError);
            this.managerName = managerName;
        }


        @Override
        protected boolean isAllFrameDataAvailable()
        {
            return currentFrameInfo != null && currentTrackingResult != null && currentPoseData != null
                    && currentFrameInfo.frameId == currentTrackingResult.getFrameId()
                    && currentFrameInfo.frameId == currentPoseData.getFrameId();
        }


        @SuppressWarnings("DuplicatedCode")
        @Override
        protected void ensureNextFrameData() throws InterruptedException
        {
            while (!isAllFrameDataAvailable())
            {
                if (currentFrameInfo == null)
                {
                    currentFrameInfo = frameQueue.take();
                }

                if (currentTrackingResult == null)
                {
                    currentTrackingResult = trackingQueue.take();
                }

                if (currentPoseData == null)
                {
                    currentPoseData = poseQueue.take();
                }

                if (currentFrameInfo.frameId != currentTrackingResult.getFrameId())
                {
                    if (currentFrameInfo.frameId == -1 || currentTrackingResult.getFrameId() != -1
                            && currentTrackingResult.getFrameId() < currentFrameInfo.frameId)
                    {
                        LOGGER.warn("Missing a frame with id {} [{}]", currentTrackingResult.getFrameId(),
                                managerName);
                        currentTrackingResult = trackingQueue.take();
                    }
                    else
                    {
                        LOGGER.warn("Missing a tracking with frame id {} [{}]", currentFrameInfo.frameId, managerName);
                        currentFrameInfo = frameQueue.take();
                    }
                }

                if (currentFrameInfo.frameId != currentPoseData.getFrameId())
                {
                    if (currentFrameInfo.frameId == -1 || currentPoseData.getFrameId() != -1
                            && currentPoseData.getFrameId() < currentFrameInfo.frameId)
                    {
                        LOGGER.warn("Missing a frame with id {} [{}]", currentPoseData.getFrameId(), managerName);
                        currentPoseData = poseQueue.take();
                    }
                    else
                    {
                        LOGGER.warn("Missing a pose estimation with frame id {} [{}]", currentFrameInfo.frameId,
                                managerName);
                        currentFrameInfo = frameQueue.take();
                    }
                }
            }
        }


        @Override
        protected long getCurrentFrameId()
        {
            return currentFrameInfo.frameId;
        }


        @Override
        protected void clearFrameData()
        {
            currentFrameInfo = null;
            currentTrackingResult = null;
            currentPoseData = null;
        }


        @Override
        protected void buildRequestPart(PerceptionRequest.Builder rb)
        {
            rb.getDistanceBuilder().setImage(ByteString.copyFrom(currentFrameInfo.frame));
            rb.getDistanceBuilder().setCameraHeightMm(cameraHeightMm);
            rb.getDistanceBuilder().setFocalLengthMm(focalLengthMm);
            rb.getDistanceBuilder().setPixelSizeMm(pixelSizeMm);

            if (normal != null && (normal.getX() != 0 || normal.getY() != 0 || normal.getZ() != 0))
            {
                rb.getDistanceBuilder().getNormalVectorBuilder().setX(normal.getX()).setY(normal.getY())
                        .setZ(normal.getZ());
            }

            Map<Long, PoseEstimationEntry> poses = currentPoseData.getPose3D().getEntriesList().stream()
                    .collect(Collectors.toMap(PoseEstimationEntry::getTrackId, p -> p));
            for (TrackingEntryOrBuilder te : currentTrackingResult.getTrackingOrBuilder().getEntriesOrBuilderList())
            {
                DistanceEstimationEntry.Builder deBuilder = rb.getDistanceBuilder().addEntriesBuilder();
                if (te.getDetection().getClassId() == DetectionClass.PERSON.getId())
                {
                    deBuilder.addAllPoints(poses.get(te.getId()).getJointsList());
                }
                else
                {
                    float x = (te.getDetection().getBoxTopLeftX() + te.getDetection().getBoxBottomRightX()) / 2.0f;
                    float y = te.getDetection().getBoxBottomRightY() - boxContactPointLiftPercent *
                            (te.getDetection().getBoxBottomRightY() - te.getDetection().getBoxTopLeftY());
                    deBuilder.addPointsBuilder().setX(x).setY(y);
                }
            }
        }
    }
}
