package com.devbeh.managers;


import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.atomic.AtomicLong;

import com.devbeh.Actor;
import com.devbeh.FrameIdAndBytes;
import com.devbeh.MetricService;
import com.devbeh.VideoInfo;
import com.devbeh.videoproviders.VideoProvider;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


/**
 * Запускает отдельный поток, видеофайл и
 */
public class VideoProvisionManager extends Actor
{
    private static final Logger LOGGER = LoggerFactory.getLogger(VideoProvisionManager.class);

    /**
     * Счетчик ошибок.
     */
    private final AtomicLong errorCounter = new AtomicLong();

    /**
     * Название видеопотока, используемое при логировании и в именах потоков.
     */
    private final String managerName;

    /**
     * Источник видео.
     */
    private final VideoProvider videoProvider;

    /**
     * Максимальный размер кадра в пикселях, полученный как height * width. В случае, если кадры видеозаписи
     * превышают этот размер, провайдер будет осуществлять ресайз.
     */
    private final int maxFrameSizePixels;

    /**
     * Должна быть инициализирована fps обрабатываемого видеопотока.
     */
    private final CompletableFuture<VideoInfo> fpsFuture;

    // Получатели кадров видеозаписи
    private final DetectionManager dm;
    private final TrackingManager tm;
    private final PoseEstimationManager pem;
    private final DistanceEstimationManager dem;
    private final PerceptionVisualizationManager pvm;

    /**
     * Может быть вызван для остановки всего pipeline в случае возникновения критичной ошибки.
     */
    private final Runnable pipelineStopper;

    /**
     * Используется для отображения метрик на странице metrics.
     */
    private final MetricService metricService;

    /**
     * Имя метрики последнего доставленного идентификатора кадра.
     */
    private final String lastFrameIdMetricName;

    /**
     * Имя метрики количества ошибок.
     */
    private final String totalErrorsMetricName;


    public VideoProvisionManager(String videoName, VideoProvider videoProvider, int maxFrameSizePixels,
            CompletableFuture<VideoInfo> fpsFuture, DetectionManager dm, TrackingManager tm,
            PoseEstimationManager pem, DistanceEstimationManager dem, PerceptionVisualizationManager pvm,
            Runnable pipelineStopper, MetricService metricService)
    {
        this.managerName = "VideoProvisionManager-" + videoName;
        this.videoProvider = videoProvider;
        this.maxFrameSizePixels = maxFrameSizePixels;
        this.fpsFuture = fpsFuture;
        this.dm = dm;
        this.tm = tm;
        this.pem = pem;
        this.dem = dem;
        this.pvm = pvm;
        this.pipelineStopper = pipelineStopper;
        this.metricService = metricService;
        this.lastFrameIdMetricName = videoName + ".provision.lastFrameId";
        this.totalErrorsMetricName = videoName + ".provision.totalErrors";

        metricService.submit(totalErrorsMetricName, 0);
    }


    private void incrementErrorCounter()
    {
        metricService.submit(totalErrorsMetricName, errorCounter.incrementAndGet());
    }


    @Override
    protected void doFirst()
    {
        try
        {
            // Инициализируем videoProvider для получения fps даже если потребителям не требуются данные, при этом
            // провайдер может определить Long.MAX_VALUE
            videoProvider.initVideoInfo(maxFrameSizePixels);
            fpsFuture.complete(new VideoInfo(videoProvider.getHeight(), videoProvider.getWidth(),
                    videoProvider.getFps()));

            long lastConsumedByAllFrameId = Math.min(
                    Math.min(
                            Math.min(dm.getInitialLastConsumedFrameId(), tm.getInitialLastConsumedFrameId()),
                            Math.min(pem.getInitialLastDistributedFrameId(), dem.getInitialLastConsumedFrameId())),
                    pvm.getInitialLastConsumedFrameId());
            long initialFrameId = lastConsumedByAllFrameId == Long.MAX_VALUE ? Long.MAX_VALUE :
                    lastConsumedByAllFrameId + 1;

            videoProvider.initInitialFrameId(initialFrameId);

            if (initialFrameId == Long.MAX_VALUE)
            {
                // Потребители уже имеют всю информацию вплоть до завершения видео
                LOGGER.info("Stopping {} without initialization. All consumers already received all frames data",
                        managerName);
                stop();
                //noinspection UnnecessaryReturnStatement
                return;
            }
        }
        catch (Exception e)
        {
            if (!getDoStop())
            {
                LOGGER.error("A fatal exception occurred during video provider initialization [{}]", managerName, e);
                incrementErrorCounter();
                // Останавливаем весь pipeline
                pipelineStopper.run();
            }
        }
    }


    @Override
    protected void process() throws InterruptedException
    {
        Optional<FrameIdAndBytes> frameInfo;
        try
        {
            // Гарантировано получение кадров в порядке возрастания
            frameInfo = videoProvider.read(200);
            if (frameInfo.isEmpty())
            {
                return;
            }
        }
        catch (Exception e)
        {
            LOGGER.error("A fatal exception occurred during reading video [{}]", managerName, e);
            incrementErrorCounter();
            // Останавливаем весь pipeline
            pipelineStopper.run();
            return;
        }

        // Если любой из вызовов завершится Interrupted исключением, выходит из данного метода без отправки
        // сообщения остальным потребителям. Предполагаем, что Interrupt может произойти только при завершении
        // работы. Но если это не так, инцидент будет залогирован, а кадр пропущен. Потребители готовы к пропуску
        // кадра. Если же пришел завершающий кадр, и Interrupt был ложным, произойдет повторная отправка
        // завершающего сообщения. К этому менеджеры также должны быть готовы.
        sendFrameIfRequired(dm, frameInfo.get());
        sendFrameIfRequired(tm, frameInfo.get());
        sendFrameIfRequired(pem, frameInfo.get());
        sendFrameIfRequired(dem, frameInfo.get());
        sendFrameIfRequired(pvm, frameInfo.get());

        if (frameInfo.get().frameId == -1)
        {
            // Достигнут конец видеопотока, останавливаем данный менеджер
            stop();
        }

        metricService.submit(lastFrameIdMetricName, frameInfo.get().frameId);
    }


    public static void sendFrameIfRequired(FrameConsumer frameConsumer, FrameIdAndBytes frameInfo)
            throws InterruptedException
    {
        if (frameInfo.frameId == -1 || frameConsumer.getInitialLastConsumedFrameId() < frameInfo.frameId)
        {
            frameConsumer.putFrame(frameInfo);
        }
    }


    @Override
    protected void doLast()
    {
        try
        {
            videoProvider.close();
        }
        catch (Exception e)
        {
            LOGGER.error("Exception occurred during video provider close [{}]", managerName);
            incrementErrorCounter();
        }
    }


    @Override
    public String getName()
    {
        return managerName;
    }


    @Override
    protected boolean isDaemon()
    {
        return false;
    }
}
