package com.devbeh.videoproviders;


import java.io.IOException;
import java.util.Optional;

import com.devbeh.FrameIdAndBytes;
import org.bytedeco.javacpp.BytePointer;
import org.bytedeco.javacv.FFmpegFrameGrabber;
import org.bytedeco.javacv.Frame;
import org.bytedeco.javacv.FrameGrabber;
import org.bytedeco.javacv.OpenCVFrameConverter;
import org.bytedeco.opencv.global.opencv_imgcodecs;
import org.bytedeco.opencv.global.opencv_imgproc;
import org.bytedeco.opencv.opencv_core.Mat;
import org.bytedeco.opencv.opencv_core.Size;


/**
 * Реализация {@link VideoProvider}, считывающая видеопоток из локального файла.
 */
public class FileVideoProvider implements VideoProvider
{
    /**
     * Путь к файлу видеозаписи.
     */
    private final String filePath;

    /**
     * Ожидаемое на выходе максимальное количество кадров в секунду.
     */
    private final int maxFps;

    /**
     * Осуществляет преобразование кадра в Mat формат.
     */
    private OpenCVFrameConverter.ToMat converterToMat;

    /**
     * Используется для получения закодированного изображения. При необходимости размер будет автоматически изменен.
     */
    private BytePointer outputPointer;

    /**
     * Номер начального ожидаемого кадра. Первый возвращенный кадр должен быть не меньше.
     */
    private long initialFrameId;

    /**
     * Используется для чтения видеозаписи.
     */
    private FrameGrabber fg;

    /**
     * Интервал между возвращаемым кадрами в миллисекундах.
     */
    private double frameIntervalMs;

    /**
     * Интервал между кадрами видеозаписи в миллисекундах.
     */
    private double videoFrameIntervalMs;

    /**
     * Требуемая измененная ширина и высота кадра в пикселях, или null.
     */
    private Size alteredSize;

    /**
     * Счетчик возвращаемых кадров.
     */
    private int frameCounter;

    /**
     * Счетчик кадров видеозаписи.
     */
    private int videoFrameCounter;

    /**
     * Ширина возвращаемого изображения.
     */
    private int width;

    /**
     * Высота возвращаемого изображения.
     */
    private int height;


    public FileVideoProvider(String filePath, int maxFps)
    {
        this.filePath = filePath;
        this.maxFps = maxFps;
    }


    @Override
    public void initVideoInfo(int maxFrameSizePixels) throws FrameGrabber.Exception
    {
        converterToMat = new OpenCVFrameConverter.ToMat();
        outputPointer = new BytePointer();

        frameCounter = 0;
        videoFrameCounter = 0;

        fg = new FFmpegFrameGrabber(filePath);
        fg.start();

        videoFrameIntervalMs = 1000.0 / fg.getFrameRate();
        frameIntervalMs = Math.max(1000.0 / maxFps, videoFrameIntervalMs);

        double frameScaleFactor = Math.sqrt((double) maxFrameSizePixels / (fg.getImageHeight() * fg.getImageWidth()));
        if (frameScaleFactor >= 1)
        {
            width = fg.getImageWidth();
            height = fg.getImageHeight();
        }
        else
        {
            width = (int) (fg.getImageWidth() * frameScaleFactor);
            height = (int) (fg.getImageHeight() * frameScaleFactor);
        }
        // Для создания нового Mat в отдельной области памяти определяем данный объект даже при неизменности размера
        // кадра
        alteredSize = new Size(width, height);
    }


    @Override
    public void initInitialFrameId(long initialFrameId)
    {
        this.initialFrameId = initialFrameId;
    }


    @Override
    public int getWidth()
    {
        return width;
    }


    @Override
    public int getHeight()
    {
        return height;
    }


    @Override
    public double getFps()
    {
        return 1000 / frameIntervalMs;
    }


    @Override
    public Optional<FrameIdAndBytes> read(long timeoutMs) throws IOException
    {
        long startTimeMs = System.currentTimeMillis();
        do {
            Frame f = fg.grabFrame();
            if (f == null)
            {
                return Optional.of(new FrameIdAndBytes(-1, null, null));
            }

            if (videoFrameCounter++ * videoFrameIntervalMs >= frameCounter * frameIntervalMs)
            {
                int frameId = frameCounter++;
                if (frameId >= initialFrameId)
                {
                    Mat frameMat = converterToMat.convert(f);
                    if (alteredSize != null)
                    {
                        Mat resizeFrameMat = new Mat();
                        opencv_imgproc.resize(frameMat, resizeFrameMat, alteredSize);
                        frameMat = resizeFrameMat;
                    }
                    opencv_imgcodecs.imencode(".jpg", frameMat, outputPointer);
                    byte[] frameBytes = new byte[(int) outputPointer.limit()];
                    outputPointer.position(0);
                    outputPointer.get(frameBytes);
                    // Следует реализовать закрытие Mat объекта
                    return Optional.of(new FrameIdAndBytes(frameId, frameBytes, frameMat));
                }
            }
        } while(System.currentTimeMillis() - startTimeMs < timeoutMs);
        return Optional.empty();
    }


    @Override
    public void close() throws FrameGrabber.Exception
    {
        if (converterToMat != null)
        {
            converterToMat.close();
            converterToMat = null;
        }

        if (outputPointer != null)
        {
            outputPointer.close();
            outputPointer = null;
        }

        if (fg != null)
        {
            fg.close();
            fg = null;
        }
    }
}
